from ..particle_tracker import ParticleTracker
import numpy as np
from ..trajectory import Trajectory
from ..frames.frames import Frames
from scipy.signal import find_peaks


class ShortestPathFinder:
    """
    Class for finding shortest path between points in a set of frames.

    Parameters
    ----------
    frames: np.array
        The frames in which trajectories are to be found. The shape of the np.array should be (nFrames,xPixels). The intensity of the frames should be normalised according to
        :math:`I_n = (I-I_{min})/(I_{max}-I_{min})`, where :math:`I` is the intensity of the frames, :math:`I_{min}`, :math:`I_{max}` are the global intensity minima and maxima of the
        frames.
    time: np.array
        The corresponding time of each frame.
    automatic_update: bool
        Choose if the class should update itself when changing properties.

    Attributes
    ----------
    frames
    time
    boxcar_width
    integration_radius_of_intensity_peaks
    particle_detection_threshold
    particle_positions
    """

    def __init__(self, frames, time, automatic_update=True):
        ParticleTracker._validate_class_arguments(frames, time, automatic_update)
        self._Frames = Frames(frames=frames, time=time, automatic_update=automatic_update)
        self._automatic_update = automatic_update
        self._integration_radius_of_intensity_peaks = 1
        self._particle_detection_threshold = 0
        self._averaged_intensity = frames
        self._particle_positions = [None] * self.frames.shape[0]
        self._cost_matrix = []
        self._association_matrix = []
        self._start_point = (0, 0)
        self._end_point = (0, 0)
        self._shortest_path = None
        self._trajectory = None

    @property
    def frames(self):
        """
        np.array:
            The frames which the particle tracker tries to find trajectories in. If the property boxcar_width!=0 it will return the smoothed frames.
        """
        return self._Frames.frames

    @property
    def boxcar_width(self):
        """
        int:
            Number of values used in the boxcar averaging of the frames.
        """
        return self._Frames.boxcar_width

    @boxcar_width.setter
    def boxcar_width(self, width):
        if not width == self._Frames.boxcar_width:
            self._Frames.boxcar_width = width
            if self._automatic_update:
                self._find_particle_positions()
                self._update_shortest_path()

    @property
    def integration_radius_of_intensity_peaks(self):
        """
        int:
            Number of pixels used when integrating the intensity peaks. No particles closer than twice this value will be found. If two peaks are found within twice this value,
            the one with highest intensity moment will be kept.
        """
        return self._integration_radius_of_intensity_peaks

    @integration_radius_of_intensity_peaks.setter
    def integration_radius_of_intensity_peaks(self, radius):
        if type(radius) is not int:
            raise TypeError('Attribute integration_radius_of_intensity_peaks should be of type int')
        if not -1 < radius <= self.frames.shape[1] / 2:
            raise ValueError('Attribute integration_radius_of_intensity_peaks should be a positive integer less or equal the half of the number of pixels in each frame.')

        if not radius == self._integration_radius_of_intensity_peaks:
            self._integration_radius_of_intensity_peaks = radius
            if self._automatic_update:
                self._find_particle_positions()
                self._update_shortest_path()

    @property
    def particle_detection_threshold(self):
        """
        float:
            Defines the threshold value for finding intensity peaks. Local maximas below this threshold will not be
            considered as particles. Should be a value between 0 and 1.
        """
        return self._particle_detection_threshold

    @particle_detection_threshold.setter
    def particle_detection_threshold(self, threshold):
        if not (type(threshold) == int or type(threshold) == float):
            raise TypeError('Attribute particle_detection_threshold should be a numerical value between 0 and 1.')
        if not 0 <= threshold <= 1:
            raise ValueError('Attribute particle_detection_threshold should be a value between 0 and 1.')
        if not threshold == self._particle_detection_threshold:
            self._particle_detection_threshold = threshold
            if self._automatic_update:
                self._find_particle_positions()
                self._update_shortest_path()

    @property
    def start_point(self):
        """
        tuple:
            (frame_index, position_index), The start point of the path you want to find.
        """
        return self._start_point

    @start_point.setter
    def start_point(self, start_point):
        if type(start_point) not in [list, tuple, np.ndarray]:
            raise TypeError('Start point must be list, tuple or np.array with form (frame_index, position_index)')
        else:
            for val in start_point:
                if type(val) not in [int, np.int16, np.int32]:
                    raise TypeError('Start point values must be integers')
            if start_point[0] < 0 or start_point[0] > self._Frames.time.shape[0] - 1:
                raise ValueError('First value in start point must be an integer between 0 and self.time.shape[0] - 1')
            elif start_point[1] < 0 or start_point[1] > self.frames.shape[1] - 1:
                raise ValueError('Second value in start point must be an integer between 0 and self.frames.shape[1] - 1')

            if not (start_point[0] == self._start_point[0] and start_point[1] == self._start_point[1]):
                self._start_point = (int(start_point[0]), int(start_point[1]))
                if self._automatic_update:
                    self._find_particle_positions()
                    self._update_shortest_path()

    @property
    def end_point(self):
        """
        tuple:
            (frame_index, position_index), The end point of the path you want to find.
        """
        return self._end_point

    @end_point.setter
    def end_point(self, end_point):
        if type(end_point) not in [list, tuple, np.ndarray]:
            raise TypeError('End point must be list, tuple or np.array with form (frame_index, position_index)')
        else:
            for val in end_point:
                if type(val) not in [int, np.int16, np.int32]:
                    raise TypeError('End point values must be integers')
            if end_point[0] < 0 or end_point[0] > self._Frames.time.shape[0] - 1:
                raise ValueError('First value in end point must be an integer between 0 and self.time.shape[0] - 1')
            elif end_point[1] < 0 or end_point[1] > self.frames.shape[1] - 1:
                raise ValueError('Second value in end point must be an integer between 0 and self.frames.shape[1] - 1')

            if not (end_point[0] == self._end_point[0] and end_point[1] == self._end_point[1]):
                self._end_point = (int(end_point[0]), int(end_point[1]))
                if self._automatic_update:
                    self._find_particle_positions()
                    self._update_shortest_path()

    @property
    def particle_positions(self):
        """
        np.array:
            Numpy array with all particle positions on the form `np.array((nParticles,), dtype=[('frame_index', np.int16),
            ('time', np.float32),('integer_position', np.int16), ('refined_position', np.float32)])`
        """
        return self._particle_positions

    @property
    def shortest_path(self):
        """
        dict:
            The shortest path between the start and end point, defined by the cost function. Cost, length and association matrix.
        """
        return self._shortest_path

    @property
    def trajectory(self):
        """
        trajectory:
            The shortest path between the start and end point represented as a trajectory.
        """
        return self._trajectory

    @property
    def _intensity_of_interest(self):
        return self.frames[self.start_point[0]:self.end_point[0] + 1]

    @property
    def _time_of_interest(self):
        return self._Frames.time[self.start_point[0]:self.end_point[0] + 1]

    def _find_particle_positions(self):
        if self.start_point and self.end_point and self.start_point[0] < self.end_point[0]:
            self._find_initial_particle_positions()
            self._refine_particle_positions()

    def _refine_particle_positions(self):
        if self._integration_radius_of_intensity_peaks == 0:
            return
        for frame_index, positions in enumerate(self._particle_positions):
            for index, position in enumerate(positions):
                if position == 0 or position + 1 == self._intensity_of_interest.shape[1]:
                    continue
                elif position < self.integration_radius_of_intensity_peaks:
                    integration_radius = position
                elif position > self._intensity_of_interest.shape[1] - self._integration_radius_of_intensity_peaks - 1:
                    integration_radius = self._intensity_of_interest.shape[1] - position
                else:
                    integration_radius = self.integration_radius_of_intensity_peaks
                intensity = self._intensity_of_interest[frame_index][int(position - integration_radius):int(position + integration_radius + 1)]
                intensity = intensity - np.min(intensity)
                self._particle_positions[frame_index][index] = position + ParticleTracker._calculate_center_of_mass(intensity) - integration_radius

    def _find_initial_particle_positions(self):
        self._particle_positions = [None] * (self.end_point[0] - self.start_point[0] + 1)
        for index, frame in enumerate(self._intensity_of_interest):
            self._particle_positions[index] = self._find_local_maximas_larger_than_threshold(frame)
        self._particle_positions[0] = np.array([self.start_point[1]], dtype=np.float32)
        self._particle_positions[-1] = np.array([self.end_point[1]], dtype=np.float32)

    def _initialise_association_and_cost_matrix(self):
        number_of_frames = len(self._particle_positions) - 1

        self._association_matrix = [[] for _ in range(number_of_frames)]
        self._cost_matrix = [[] for _ in range(number_of_frames)]

        for frame_index in range(0, number_of_frames):
            self._association_matrix[frame_index] = np.zeros(
                (len(self._particle_positions[frame_index]), len(self._particle_positions[frame_index + 1])),
                dtype=bool
            )
            self._cost_matrix[frame_index] = np.zeros(
                (len(self._particle_positions[frame_index]), len(self._particle_positions[frame_index + 1])),
                dtype=np.float32
            )

    def _calculate_cost_matrix(self):
        for frame_index, _ in enumerate(self._cost_matrix):
            for particle_index, _ in enumerate(self._cost_matrix[frame_index]):
                for future_particle_index, _ in enumerate(self._cost_matrix[frame_index][particle_index]):
                    if frame_index == 0:
                        self._cost_matrix[frame_index][particle_index][future_particle_index] = self._calculate_linking_cost(
                            frame_index,
                            particle_index,
                            frame_index + 1,
                            future_particle_index
                        )
                    else:
                        self._cost_matrix[frame_index][particle_index][future_particle_index] = (
                            self._calculate_linking_cost(
                                frame_index,
                                particle_index,
                                frame_index + 1,
                                future_particle_index
                            ) +
                            np.amin(self._cost_matrix[frame_index - 1].T[particle_index])
                        )

    def _calculate_linking_cost(self, frame_index, particle_index, future_frame_index, future_particle_index):
        return (
                0.1*(self.particle_positions[frame_index][particle_index] - self.particle_positions[future_frame_index][future_particle_index]) ** 2 +
                (self._zeroth_order_moments[future_frame_index][future_particle_index] - self._zeroth_order_moments[0][0]) ** 2
        )

    def _update_shortest_path(self):
        if self.start_point and self.end_point and self.start_point[0] < self.end_point[0]:
            self._calculate_particle_moments()
            self._initialise_association_and_cost_matrix()
            self._calculate_cost_matrix()
            self._create_trajectory_from_cost_matrix()

    def _calculate_particle_moments(self):
        self._zeroth_order_moments = [None] * (self.end_point[0] - self.start_point[0] + 1)
        self._second_order_moments = [None] * (self.end_point[0] - self.start_point[0] + 1)
        for frame_index, positions in enumerate(self._particle_positions):
            self._zeroth_order_moments[frame_index] = np.array([self._calculate_zeroth_order_intensity_moment(position, frame_index + self.start_point[0]) for position in positions], dtype=np.float64)
            self._second_order_moments[frame_index] = np.array([self._calculate_second_order_intensity_moment(position, frame_index + self.start_point[0]) for position in positions], dtype=np.float64)

    def _create_trajectory_from_cost_matrix(self):
        particle_indices = []
        prev_lowest_cost_index = 0
        for frame_index in range(len(self._cost_matrix)):
            lowest_cost_index = np.where(
                self._cost_matrix[-frame_index - 1].T[prev_lowest_cost_index] == np.amin(self._cost_matrix[-frame_index - 1].T[prev_lowest_cost_index])
            )[0][0]
            particle_indices.append(
                lowest_cost_index
            )
            prev_lowest_cost_index = lowest_cost_index

        particle_positions = np.empty((len(particle_indices[::-1]),), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32), ('second_order_moment', np.float32)])
        self._trajectory = Trajectory()
        for frame_index, particle_index in enumerate(particle_indices[::-1]):
            particle_positions[frame_index]['frame_index'] = frame_index + self.start_point[0]
            particle_positions[frame_index]['time'] = self._Frames.time[frame_index + self.start_point[0]]
            particle_positions[frame_index]['position'] = self._particle_positions[frame_index][particle_index]
            particle_positions[frame_index]['zeroth_order_moment'] = self._zeroth_order_moments[frame_index][particle_index]
            particle_positions[frame_index]['second_order_moment'] = self._second_order_moments[frame_index][particle_index]

        self._trajectory.particle_positions = particle_positions

    def _calculate_second_order_intensity_moment(self, position, frame_index):
        position = int(round(position))
        if self._integration_radius_of_intensity_peaks == 0:
            return 0
        elif position == 0:
            if self._integration_radius_of_intensity_peaks == 1:
                return 2 * self.frames[frame_index, 1] / self._calculate_zeroth_order_intensity_moment(position, frame_index)
            else:
                second_order_index_array = np.arange(0, self._integration_radius_of_intensity_peaks + 1) ** 2
                return (
                               2 * np.dot(self.frames[frame_index, :self.integration_radius_of_intensity_peaks + 1], second_order_index_array)
                       ) / self._calculate_zeroth_order_intensity_moment(position, frame_index)
        elif position == self.frames.shape[1] - 1:
            if self._integration_radius_of_intensity_peaks == 1:
                return 2 * self.frames[frame_index, -2] / self._calculate_zeroth_order_intensity_moment(position, frame_index)
            else:
                second_order_index_array = np.arange(-self._integration_radius_of_intensity_peaks, 0) ** 2
                return 2 * np.dot(self.frames[frame_index, -self._integration_radius_of_intensity_peaks - 1:-1],
                                  second_order_index_array) / self._calculate_zeroth_order_intensity_moment(position, frame_index)
        elif position < self._integration_radius_of_intensity_peaks:
            w = self._integration_radius_of_intensity_peaks - position
            if w == 1:
                second_order_index_array = np.arange(-position, position + 1) ** 2
                return (
                        (
                                np.dot(self.frames[frame_index, :2 * position + 1], second_order_index_array) +
                                2 * self._integration_radius_of_intensity_peaks ** 2 * self.frames[frame_index, position + self._integration_radius_of_intensity_peaks]
                        ) / self._calculate_zeroth_order_intensity_moment(position, frame_index)
                )
            else:
                second_order_index_array = np.arange(-position, position + 1) ** 2
                second_order_index_array_big = np.arange(position + 1, position + self._integration_radius_of_intensity_peaks) ** 2
                return (
                               np.dot(second_order_index_array, self.frames[frame_index, :2 * position + 1]) +
                               2 * np.dot(second_order_index_array_big,
                                          self.frames[frame_index, 2 * position + 1:2 * position + self._integration_radius_of_intensity_peaks])
                       ) / self._calculate_zeroth_order_intensity_moment(position, frame_index)
        elif position > self.frames.shape[1] - 1 - self._integration_radius_of_intensity_peaks:
            w = self._integration_radius_of_intensity_peaks - (self.frames.shape[1] - position - 1)
            if w == 1:
                second_order_index_array = np.arange(-(self.frames.shape[1] - 1 - position), self.frames.shape[1] - position) ** 2
                return (
                               np.dot(second_order_index_array, self.frames[frame_index, 2 * position - self.frames.shape[1] + 1:])
                               + 2 * self._integration_radius_of_intensity_peaks ** 2 * self.frames[
                                   frame_index, position - self._integration_radius_of_intensity_peaks]
                       ) / self._calculate_zeroth_order_intensity_moment(position, frame_index)
            else:
                second_order_index_array = np.arange(-(self.frames.shape[1] - 1 - position), self.frames.shape[1] - position) ** 2
                second_order_index_array_big = np.arange(- self._integration_radius_of_intensity_peaks, -(self.frames.shape[1] - position) + 1) ** 2
                return (
                               np.dot(second_order_index_array, self.frames[frame_index, -2 * (self.frames.shape[1] - position) + 1:]) +
                               2 * np.dot(second_order_index_array_big,
                                          self.frames[frame_index, position - self._integration_radius_of_intensity_peaks:-2 * (self.frames.shape[1] - position) + 1])
                       ) / self._calculate_zeroth_order_intensity_moment(position, frame_index)

        else:
            w = self._integration_radius_of_intensity_peaks
            second_order_index_array = np.arange(-w, w + 1) ** 2
            return np.dot(self.frames[frame_index, position - w:position + w + 1], second_order_index_array) / self._calculate_zeroth_order_intensity_moment(position,
                                                                                                                                                                         frame_index)

    def _calculate_zeroth_order_intensity_moment(self, position, frame_index):
        position = int(round(position))
        if self._integration_radius_of_intensity_peaks == 0:
            return self.frames[frame_index, position]
        elif position == 0:
            if self._integration_radius_of_intensity_peaks == 1:
                return self.frames[frame_index, position] + 2 * self.frames[frame_index, position + 1]
            else:
                return (
                        self.frames[frame_index, position] +
                        2 * np.sum(self.frames[frame_index, position + 1:position + 1 + self._integration_radius_of_intensity_peaks])
                )
        elif position == self.frames.shape[1] - 1:
            if self._integration_radius_of_intensity_peaks == 1:
                return self.frames[frame_index, position] + 2 * self.frames[frame_index, position - 1]
            else:
                return (
                        self.frames[frame_index, position] +
                        2 * np.sum(self.frames[frame_index, position - self._integration_radius_of_intensity_peaks:position])
                )
        elif position < self._integration_radius_of_intensity_peaks:
            w = self._integration_radius_of_intensity_peaks - position
            if w == 1:
                return np.sum(self.frames[frame_index, :2 * position + 1]) + 2 * self.frames[frame_index, position + w + 1]
            else:
                return (
                        np.sum(self.frames[frame_index, :2 * position + 1]) +
                        2 * np.sum(self.frames[frame_index, 2 * position + 1:2 * position + w + 1])
                )
        elif position > self.frames.shape[1] - 1 - self._integration_radius_of_intensity_peaks:
            w = self._integration_radius_of_intensity_peaks - (self.frames.shape[1] - position - 1)
            if w == 1:
                return (
                        np.sum(self.frames[frame_index, -2 * (self.frames.shape[1] - position) + 1:]) +
                        2 * self.frames[frame_index, -2 * (self.frames.shape[1] - position)]
                )
            else:
                return (
                        np.sum(self.frames[frame_index, -2 * (self.frames.shape[1] - position) + 1:]) +
                        2 * np.sum(self.frames[frame_index, position - self._integration_radius_of_intensity_peaks:position - 1])
                )
        else:
            w = self._integration_radius_of_intensity_peaks
            return np.sum(self.frames[frame_index, position - w:position + w + 1])

    def _find_local_maximas_larger_than_threshold(self, y):
        local_maximas, _ = find_peaks(y, height=self.particle_detection_threshold, distance=2 * self.integration_radius_of_intensity_peaks)
        return local_maximas.astype(np.float32)

    @staticmethod
    def _validate_class_arguments(frames, time, automatic_update):
        ParticleTracker._test_if_frames_have_correct_format(frames)
        ParticleTracker._test_if_time_has_correct_format(time)
        ParticleTracker._test_if_time_and_frames_has_same_length(time, frames)
        ParticleTracker._test_if_automatic_update_has_correct_format(automatic_update)

    @staticmethod
    def _test_if_frames_have_correct_format(frames):
        if type(frames) is not np.ndarray:
            raise TypeError('Class argument frames not of type np.ndarray')
        if not (len(frames.shape) == 2 and frames.shape[0] > 1 and frames.shape[1] > 2):
            raise ValueError('Class argument frames need to be of shape (nFrames,nPixels) with nFrames > 1 and nPixels >2')
        if not (np.max(frames.flatten()) == 1 and np.min(frames.flatten()) == 0):
            raise ValueError('Class argument frames not normalised. Max value of frames should be 1 and min value should be 0.')

        return True

    @staticmethod
    def _test_if_time_has_correct_format(time):
        if type(time) is not np.ndarray:
            raise TypeError('Class argument frames not of type np.ndarray')
        if not (len(time.shape) == 1 and time.shape[0] > 1):
            raise ValueError('Class argument time need to be of shape (nFrames,) with nFrames > 1.')
        if not all(np.diff(time) > 0):
            raise ValueError('Class argument time not increasing monotonically.')
        return True

    @staticmethod
    def _test_if_time_and_frames_has_same_length(time, frames):
        if not time.shape[0] == frames.shape[0]:
            raise ValueError('Class arguments time and frames does not of equal length.')
        return True

    @staticmethod
    def _test_if_automatic_update_has_correct_format(automatic_update):
        if not type(automatic_update) == bool:
            raise ValueError('Class argument automatic_update must be True or False.')
        return True



import numpy as np


class Hindrance:

    def __init__(self):
        self._channel_x_dimension = None
        self._channel_y_dimension = None
        self._molecule_radius = None

    @property
    def channel_x_dimension(self):
        return self._channel_x_dimension

    @channel_x_dimension.setter
    def channel_x_dimension(self, dimension):
        self._channel_x_dimension = dimension

    @property
    def channel_y_dimension(self):
        return self._channel_y_dimension

    @channel_y_dimension.setter
    def channel_y_dimension(self, dimension):
        self._channel_y_dimension = dimension

    @property
    def molecule_radius(self):
        return self._molecule_radius

    @molecule_radius.setter
    def molecule_radius(self, radius):
        self._molecule_radius = radius

    def _calculate_equilibrium_partition_coefficient(self):
        if self._channel_x_dimension and self._channel_y_dimension and self._molecule_radius:
            return (1 - self._molecule_radius / np.sqrt(self._channel_x_dimension * self._channel_y_dimension))**2
        else:
            raise ValueError('Channel dimensions or molecule radius has not been set.')

    def calculate_hindrance_factor(self):
        equilibrium_partition_coefficient = self._calculate_equilibrium_partition_coefficient()
        ratio_molecule_size_and_dimension = self._molecule_radius / np.sqrt(self._channel_x_dimension * self._channel_y_dimension)
        H = 1 + 9 / 8 * ratio_molecule_size_and_dimension * np.log(ratio_molecule_size_and_dimension) - 1.56034 * ratio_molecule_size_and_dimension + \
            0.528155 * ratio_molecule_size_and_dimension ** 2 + 1.91521 * ratio_molecule_size_and_dimension ** 3 - \
            2.81903 * ratio_molecule_size_and_dimension ** 4 + 0.270788 * ratio_molecule_size_and_dimension ** 5 + \
            1.10115 * ratio_molecule_size_and_dimension ** 6 - 0.435933 * ratio_molecule_size_and_dimension ** 7
        return H / equilibrium_partition_coefficient



import unittest
import numpy as np
from particle_tracker_one_d import Trajectory


class CalculateDiffusionCoefficientTester(unittest.TestCase):

    def test_calculating_diffusion_coefficient_using_covariance_based_method(self):
        """
        Test that the calculation of diffusion coefficient from covariance based method is correct.
        """

        # Simplest test, diffusion coefficient = 0
        expected_diffusion_coefficient = 0.0

        particle_positions = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        particle_positions['frame_index'] = [0, 1, 2]
        particle_positions['time'] = [0, 1, 2]
        particle_positions['position'] = [0, 0, 0]

        R = 1 / 4  # Representing exposure time equal to acquisition time

        pixel_width = 1

        t = Trajectory(pixel_width=pixel_width)

        t.particle_positions = particle_positions
        calculated_diffusion_coefficient, error = t.calculate_diffusion_coefficient_using_covariance_based_estimator(R=R)

        self.assertEqual(expected_diffusion_coefficient, calculated_diffusion_coefficient)


class PropertyTester(unittest.TestCase):

    def test_length_of_trajectory(self):
        """
        Test that the class property returns the length of the trajectory
        """

        particle_positions = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                   ('second_order_moment', np.float32)])

        particle_positions['frame_index'] = [0, 1, 2]
        particle_positions['time'] = [0, 1, 2]
        particle_positions['position'] = [0, 0, 0]
        particle_positions['zeroth_order_moment'] = [0, 0, 0]
        particle_positions['second_order_moment'] = [0, 0, 0]

        t = Trajectory()

        expected_length_empty_trajectory = 0
        self.assertEqual(expected_length_empty_trajectory, t.length)

        t.particle_positions = particle_positions
        expected_length = 3
        self.assertEqual(expected_length, t.length)

    def test_density_of_trajectory(self):
        """
        Test that the density property returns the correct value.
        """

        t = Trajectory()

        expected_density_empty_trajectory = 1
        self.assertEqual(expected_density_empty_trajectory, t.density)

        particle_positions = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                   ('second_order_moment', np.float32)])

        particle_positions['frame_index'] = [0, 1, 3]
        particle_positions['time'] = [0, 1, 3]
        particle_positions['position'] = [0, 0, 0]
        particle_positions['zeroth_order_moment'] = [0, 0, 0]
        particle_positions['second_order_moment'] = [0, 0, 0]

        t.particle_positions = particle_positions
        expected_density = 0.75
        self.assertEqual(expected_density, t.density)

        t.particle_positions['frame_index'] = [0, 1, 2]
        expected_density = 1
        self.assertEqual(expected_density, t.density)

        particle_positions = np.empty((5,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                   ('second_order_moment', np.float32)])

        particle_positions['frame_index'] = [1, 2, 3, 5, 7]
        particle_positions['time'] = [0, 1, 3, 4, 5]
        particle_positions['position'] = [0, 0, 0, 0, 0]
        particle_positions['zeroth_order_moment'] = [0, 0, 0, 0, 0]
        particle_positions['second_order_moment'] = [0, 0, 0, 0, 0]

        t.particle_positions = particle_positions
        expected_density = 0.7142857142857143
        self.assertEqual(expected_density, t.density)


class FunctionsTester(unittest.TestCase):

    def test_overlaps_function(self):
        t_1 = Trajectory()
        t_2 = Trajectory()

        self.assertTrue(not t_1.overlaps_with(t_2))
        self.assertTrue(not t_2.overlaps_with(t_1))

        particle_positions = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                   ('second_order_moment', np.float32)])

        particle_positions['frame_index'] = [0, 1, 3]
        particle_positions['time'] = [0, 1, 3]
        particle_positions['position'] = [0, 0, 0]
        particle_positions['zeroth_order_moment'] = [0, 0, 0]
        particle_positions['second_order_moment'] = [0, 0, 0]

        t_1.particle_positions = particle_positions

        self.assertTrue(not t_1.overlaps_with(t_2))
        self.assertTrue(not t_2.overlaps_with(t_1))

        particle_positions = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                   ('second_order_moment', np.float32)])

        particle_positions['frame_index'] = [0, 2, 3]
        particle_positions['time'] = [0, 1, 3]
        particle_positions['position'] = [0, 0, 0]
        particle_positions['zeroth_order_moment'] = [0, 0, 0]
        particle_positions['second_order_moment'] = [0, 0, 0]

        t_2.particle_positions = particle_positions

        self.assertTrue(t_1.overlaps_with(t_2))
        self.assertTrue(t_2.overlaps_with(t_1))

        particle_positions = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                   ('second_order_moment', np.float32)])

        particle_positions['frame_index'] = [10, 11, 33]
        particle_positions['time'] = [0, 1, 3]
        particle_positions['position'] = [0, 0, 0]
        particle_positions['zeroth_order_moment'] = [0, 0, 0]
        particle_positions['second_order_moment'] = [0, 0, 0]

        t_2.particle_positions = particle_positions

        self.assertTrue(not t_1.overlaps_with(t_2))
        self.assertTrue(not t_2.overlaps_with(t_1))

    def test_split_trajectories_function(self):
        """
        Test that the splitting of two trajectories works.
        """

        t1 = Trajectory()
        t2 = Trajectory()

        particle_positions1 = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                    ('second_order_moment', np.float32)])
        particle_positions1['frame_index'] = [0, 1, 2]
        particle_positions1['time'] = [0, 0, 0]
        particle_positions1['position'] = [0, 0, 0]
        particle_positions1['zeroth_order_moment'] = [0, 0, 0]
        particle_positions1['second_order_moment'] = [0, 0, 0]

        particle_positions2 = np.empty((4,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                    ('second_order_moment', np.float32)])
        particle_positions2['frame_index'] = [0, 2, 3, 4]
        particle_positions2['time'] = [0, 0, 0, 0]
        particle_positions2['position'] = [0, 0, 0, 0]
        particle_positions2['zeroth_order_moment'] = [0, 0, 0, 0]
        particle_positions2['second_order_moment'] = [0, 0, 0, 0]

        t1.particle_positions = particle_positions1
        t2.particle_positions = particle_positions2

        expected_trajectory_1 = Trajectory()
        expected_trajectory_2 = Trajectory()
        expected_trajectory_3 = Trajectory()
        expected_trajectory_4 = Trajectory()

        expected_trajectory_1.particle_positions = particle_positions1[:1]
        expected_trajectory_2.particle_positions = particle_positions1[1:2]
        expected_trajectory_3.particle_positions = particle_positions1[2:3]
        expected_trajectory_4.particle_positions = particle_positions2[2:]

        split_trajectories = t1.split(t2)

        np.testing.assert_array_equal(expected_trajectory_1.particle_positions, split_trajectories[0].particle_positions)
        np.testing.assert_array_equal(expected_trajectory_2.particle_positions, split_trajectories[1].particle_positions)
        np.testing.assert_array_equal(expected_trajectory_3.particle_positions, split_trajectories[2].particle_positions)
        np.testing.assert_array_equal(expected_trajectory_4.particle_positions, split_trajectories[3].particle_positions)

        # Change order
        t1.particle_positions = particle_positions1
        t2.particle_positions = particle_positions2

        split_trajectories = t2.split(t1)

        np.testing.assert_array_equal(expected_trajectory_1.particle_positions, split_trajectories[0].particle_positions)
        np.testing.assert_array_equal(expected_trajectory_2.particle_positions, split_trajectories[1].particle_positions)
        np.testing.assert_array_equal(expected_trajectory_3.particle_positions, split_trajectories[2].particle_positions)
        np.testing.assert_array_equal(expected_trajectory_4.particle_positions, split_trajectories[3].particle_positions)

        # Test not overlapping trajectories

        particle_positions1 = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                    ('second_order_moment', np.float32)])
        particle_positions1['frame_index'] = [0, 1, 2]
        particle_positions1['time'] = [0, 0, 0]
        particle_positions1['position'] = [0, 0, 0]
        particle_positions1['zeroth_order_moment'] = [0, 0, 0]
        particle_positions1['second_order_moment'] = [0, 0, 0]

        particle_positions2 = np.empty((4,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                    ('second_order_moment', np.float32)])
        particle_positions2['frame_index'] = [3, 4, 5, 6]
        particle_positions2['time'] = [0, 0, 0, 0]
        particle_positions2['position'] = [0, 0, 0, 0]
        particle_positions2['zeroth_order_moment'] = [0, 0, 0, 0]
        particle_positions2['second_order_moment'] = [0, 0, 0, 0]

        t1.particle_positions = particle_positions1
        t2.particle_positions = particle_positions2

        expected_trajectory_1 = Trajectory()
        expected_trajectory_2 = Trajectory()

        expected_trajectory_1.particle_positions = particle_positions1
        expected_trajectory_2.particle_positions = particle_positions2

        new_trajectories = t2.split(t1)

        np.testing.assert_array_equal(new_trajectories[1].particle_positions, expected_trajectory_1.particle_positions)
        np.testing.assert_array_equal(new_trajectories[0].particle_positions, expected_trajectory_2.particle_positions)



import unittest
import numpy as np
from particle_tracker_one_d import Frames


class ValidationOfArgumentsTester(unittest.TestCase):

    def test_validation_of_frames(self):
        """
        Test that the frames are a numpy array with shape (nFrames,nPixels).
        """

        valid_frames = [
            np.random.rand(1, 2),
            np.zeros((1, 2), dtype=np.float32),
            np.zeros((1, 21), dtype=np.int8),
            np.zeros((13, 2), dtype=np.int16),
            np.zeros((1, 20), dtype=np.float64),
            np.random.rand(2, 1),
            np.random.rand(20, 100)
        ]

        non_valid_shape_frames = [
            np.random.rand(1, 1, 1),
            np.random.rand(0, 1),
            np.random.rand(1, 0),
            np.random.rand(1, 0, 3, 4),
        ]

        non_valid_type_of_frames = [
            2,
            [[0, 1, 0], [0, 0, 0]],
            'string'
        ]

        for frames in valid_frames:
            self.assertTrue(Frames._test_if_frames_have_correct_format(frames), msg='Valid frames not accepted, frames: ' + np.array_str(frames))

        for frames in non_valid_shape_frames:
            with self.assertRaises(ValueError):
                Frames._test_if_frames_have_correct_format(frames)

        for frames in non_valid_type_of_frames:
            with self.assertRaises(TypeError):
                Frames._test_if_frames_have_correct_format(frames)

    def test_validation_time_argument(self):
        """
        Test that the validation of the class argument works. The time should be a numpy.ndarray with monotonically increasing values.
        """
        valid_times = [
            np.array([0, 1, 2, 4, 6], dtype=np.float32),
            np.array([0, 1000], dtype=np.int16),
            np.array([0.1, 0.2, 0.4], dtype=np.float32),
            np.array([1], dtype=np.float32)
        ]

        non_valid_shape_or_values_times = [
            np.array([0, 0], dtype=np.float32),
            np.array([0, -1], dtype=np.float32),
            np.array([[0, 2], [3, 5]], dtype=np.float32)
        ]

        non_valid_types_of_times = [
            '1,2,3,4',
            [1, 2, 3, 4],
            1,
            {'1': 1, '2': 2}
        ]

        for times in valid_times:
            self.assertTrue(Frames._test_if_time_has_correct_format(times), msg='Valid times not accepted, times: ' + np.array_str(times))

        for times in non_valid_shape_or_values_times:
            with self.assertRaises(ValueError):
                Frames._test_if_time_has_correct_format(times)

        for times in non_valid_types_of_times:
            with self.assertRaises(TypeError):
                Frames._test_if_time_has_correct_format(times)


class ImageRestorationTester(unittest.TestCase):

    def test_boxcar_averaging(self):
        """
        Test that the boxcar average works as it should.
        """

        frames = np.array([
            [1, 1, 1, 1, 1, 1],
            [0, 1, 0, 1, 0, 1],
            [0, 0, 1, 0, 0, 1],
            [0, 1, 2, 3, 4, 5]
        ], dtype=np.float32)

        time = np.array([0, 1, 2, 3], dtype=np.float32)

        boxcar_widths = [0, 1, 2, 3]

        expected_averaged_frames = [
            np.array([
                [1, 1, 1, 1, 1, 1],
                [0, 1, 0, 1, 0, 1],
                [0, 0, 1, 0, 0, 1],
                [0, 1, 2, 3, 4, 5]
            ], dtype=np.float32),
            np.array([
                [1, 1, 1, 1, 1, 1],
                [0, 1, 0, 1, 0, 1],
                [0, 0, 1, 0, 0, 1],
                [0, 1, 2, 3, 4, 5]
            ], dtype=np.float32),
            np.array([
                [1, 1, 1, 1, 1],
                [0.5, 0.5, 0.5, 0.5, 0.5],
                [0, 0.5, 0.5, 0, 0.5],
                [0.5, 1.5, 2.5, 3.5, 4.5]
            ], dtype=np.float32),
            np.array([
                [1, 1, 1, 1],
                [0.33333333, 0.666666667, 0.33333333, 0.66666667],
                [0.33333333, 0.33333333, 0.33333333, 0.33333333],
                [1, 2, 3, 4]
            ], dtype=np.float32)
        ]

        frames = Frames(frames=frames, time=time, automatic_update=True)

        for index, boxcar_width in enumerate(boxcar_widths):
            frames.boxcar_width = boxcar_width
            np.testing.assert_array_equal(frames.frames, expected_averaged_frames[index])



import numpy as np
import matplotlib.pyplot as plt


class Trajectory:
    """
    Object that describes a trajectory. With functions for checking if the trajectory describes real diffusion,
    convenient plotting and calculations of diffusion coefficients.

    Parameters
    ----------
    pixel_width: float
        Defines the length one pixel corresponds to. This value will be used when calculating diffusion
        coefficients. Default is 1.

    Attributes
    ----------
    pixel_width
    particle_positions: np.array
        Numpy array with all particle positions in the trajectory on the form `np.array((nParticles,), dtype=[('frame_index', np.int16),
        ('time', np.float32),('position', np.int16),('zeroth_order_moment', np.float32),('second_order_moment', np.float32)])`
    """

    def __init__(self, pixel_width=1):
        self.particle_positions = np.empty((0,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32),
                                                        ('second_order_moment', np.float32)])
        self.pixel_width = pixel_width

    def __add__(self, other):
        new_trajectory = Trajectory(pixel_width=self.pixel_width)
        if self.pixel_width != other.pixel_width:
            raise ValueError('Pixel width must be equal when adding trajectories together.')
        elif self.particle_positions.shape[0] == 0 and other._particle_positions.shape[0] == 0:
            return new_trajectory
        elif self.particle_positions.shape == (0,):
            new_trajectory.particle_positions = other._particle_positions
        elif other._particle_positions.shape == (0,):
            new_trajectory.particle_positions = self.particle_positions
        elif other._particle_positions[0]['frame_index'] == self.particle_positions[0]['frame_index']:
            raise ValueError('Both trajectories cant start at same frame index.')
        elif other._particle_positions.shape[0] == 1 and self.particle_positions.shape[0] == 1:
            if other._particle_positions['frame_index'][0] < self.particle_positions['frame_index'][0]:
                new_trajectory.particle_positions = np.append(other._particle_positions, self.particle_positions)
            else:
                new_trajectory.particle_positions = np.append(self.particle_positions, other._particle_positions)
        elif other._particle_positions['frame_index'][0] < self.particle_positions['frame_index'][0]:
            index = np.where(
                other._particle_positions['frame_index'] < self.particle_positions[0]['frame_index']
            )
            new_trajectory.particle_positions = np.append(other._particle_positions[index], self.particle_positions)
        elif self.particle_positions['frame_index'][0] < other._particle_positions['frame_index'][0]:
            index = np.where(
                self.particle_positions['frame_index'] < other._particle_positions[0]['frame_index']
            )
            new_trajectory.particle_positions = np.append(self.particle_positions[index], other._particle_positions)

        return new_trajectory

    @property
    def velocities(self):
        """
        np.array:
            The velocities the particle moves at.
        """
        if self.length > 1:
            return self._calculate_particle_velocities()
        return 0

    @property
    def density(self):
        """
        float:
            How dense the trajectory is in time. Returns self.length/(self.particle_positions['frame_index'][-1]-self.particle_positions['frame_index'][0]).
        """
        if self.length == 0 or self.length == 1:
            return 1
        return self.length / (1 + self.particle_positions['frame_index'][-1] - self.particle_positions['frame_index'][0])

    @property
    def length(self):
        """
        int:
            The length of the trajectory. Returns self.particle_postions.shape[0]
        """
        return self.particle_positions.shape[0]

    def overlaps_with(self, trajectory):
        """
        Check if the trajectories overlaps

        trajectory: Trajectory to compare with. If both trajectories has any identical elements will return true otherwise false.

        Returns
        -------
            bool
        """
        if self.length == 0 or trajectory.length == 0:
            return False
        for p in trajectory.particle_positions:
            for p2 in self.particle_positions:
                if (p['frame_index'] == p2['frame_index']) and (p['position'] == p2['position']):
                    return True
        return False

    def split(self, trajectory):
        """
        If two trajectories overlaps, this function will split them into three or more non overlapping trajectories.

        trajectory:

        Returns
        -------
            list
                Returns a list with the new trajectories
        """
        if (
                (self.particle_positions[0]['frame_index'] > trajectory.particle_positions[-1]['frame_index']) or
                (self.particle_positions[-1]['frame_index'] < trajectory.particle_positions[0]['frame_index'])
        ):
            return [self, trajectory]

        p1 = self.particle_positions.copy()
        p2 = trajectory.particle_positions.copy()

        new_positions = []
        while (p1.shape[0] > 0) or (p2.shape[0] > 0):
            if (p1.shape[0] == 0) and (p2.shape[0] > 0):
                new_positions.append(p2.copy())
                break
            elif (p2.shape[0] == 0) and (p1.shape[0] > 0):
                new_positions.append(p1.copy())
                break
            elif p1[0] == p2[0]:
                index = self._find_index_of_first_not_equal_element(p1, p2)
                new_positions.append(p1[:index].copy())
                p1 = p1[index:]
                p2 = p2[index:]
            else:
                index1, index2 = self._find_last_index_where_no_overlaps_occurs(p1, p2)
                new_positions.append(p1[:index1].copy())
                new_positions.append(p2[:index2].copy())
                p1 = p1[index1:]
                p2 = p2[index2:]

        new_trajectories = []
        for p in new_positions:
            if p.shape[0] > 0:
                t = Trajectory(pixel_width=self.pixel_width)
                t.particle_positions = p
                new_trajectories.append(t)

        return new_trajectories

    def plot_trajectory(self, x='frame_index', y='position', ax=None, **kwargs):
        """
        Plots the trajectory using the frame index and the particle position in pixels.

        x: str
            'frame_index', 'time', 'position' (default), 'zeroth_order_moment', 'second_order_moment' choose the x-axis value
        y: str
            'frame_index' (default), 'time', 'position', 'zeroth_order_moment', 'second_order_moment' choose the y-axis value
        ax: matplotlib axes instance
            The axes which you want the frames to plotted on. If none is provided a new instance will be created.
        **kwargs:
            Plot settings, any settings which can be used in matplotlib.pyplot.plot method.

        Returns
        -------
            matplotlib axes instance
                Returns the axes input argument or creates and returns a new instance of a matplotlib axes object.
        """
        if ax is None:
            ax = plt.axes()
        ax.plot(self.particle_positions[x], self.particle_positions[y], **kwargs)
        return ax

    def plot_velocity_auto_correlation(self, ax=None, **kwargs):
        """
        Plots the particle velocity auto correlation function which can be used for examining if the trajectory
        describes free diffusion.

        ax: matplotlib axes instance
            The axes which you want the frames to plotted on. If none is provided a new instance will be created.
        **kwargs:
            Plot settings, any settings which can be used in matplotlib.pyplot.plot method.

        Returns
        -------
            matplotlib axes instance
                Returns the axes input argument or creates and returns a new instance of a matplotlib axes object.
        """
        if ax is None:
            ax = plt.axes()
        ax.acorr(self.velocities, **kwargs)
        return ax

    def calculate_mean_square_displacement_function(self):
        """
        Calculate the average squared displacements for different time steps.

        Returns
        -------
            time: np.array
                The time corresponding to the mean squared displacements.

            msd: np.array
                The mean squared displacements of the trajectory.
        """

        mean_square_displacements = np.zeros((self.particle_positions['frame_index'][-1] - self.particle_positions['frame_index'][0] + 1,),
                                             dtype=[('msd', np.float32), ('nr_of_values', np.int16)])
        times = np.arange(0, self.particle_positions['frame_index'][-1] - self.particle_positions['frame_index'][0] + 1, dtype=np.float32) * self._calculate_time_step()

        for first_index, first_position in enumerate(self.particle_positions[:-1]):
            for second_index, second_position in enumerate(self.particle_positions[first_index + 1:]):
                index_difference = second_position['frame_index'] - first_position['frame_index']
                mean_square_displacements['msd'][index_difference] += ((second_position['position'] - first_position['position']) * self.pixel_width) ** 2
                mean_square_displacements['nr_of_values'][index_difference] += 1

        for index, msd in enumerate(mean_square_displacements):
            if mean_square_displacements['nr_of_values'][index] != 0:
                mean_square_displacements['msd'][index] = msd['msd'] / mean_square_displacements['nr_of_values'][index].astype(np.float32)

        non_zeros_indices = np.nonzero(mean_square_displacements['nr_of_values'])
        return times[non_zeros_indices], mean_square_displacements['msd'][non_zeros_indices]

    def _append_position(self, particle_position):
        self.particle_positions = np.append(self.particle_positions, particle_position, axis=0)

    def _position_exists_in_trajectory(self, particle_position):
        for p in self.particle_positions:
            if np.array_equal(p, particle_position):
                return True

    def _calculate_particle_velocities(self):
        time_steps = np.diff(self.particle_positions['time'])
        position_steps = np.diff(self.particle_positions['position'] * self.pixel_width)
        return position_steps / time_steps

    @staticmethod
    def _remove_non_unique_values(array):
        return np.unique(array)

    @staticmethod
    def _sort_values_low_to_high(array):
        return np.sort(array)

    def calculate_diffusion_coefficient_from_mean_square_displacement_function(self, fit_range=None):
        """
        Fits a straight line to the mean square displacement function and calculates the diffusion coefficient from the
        gradient of the line. The mean squared displacement of the particle position is proportional to :math:`2Dt`
        where :math:`D` is the diffusion coefficient and :math:`t` is the time.

        fit_range: list, None (default)
            Define the range of the fit, the data for the fit will be `time[fit_range[0]:fit_range[1]`` and `mean_squared_displacement[fit_range[0]:fit_range[1]]`.

        Returns
        -------
            diffusion_coefficient: float
            error: float
        """
        time, mean_square_displacement = self.calculate_mean_square_displacement_function()
        if fit_range is None:
            polynomial_coefficients, error_estimate = self._fit_straight_line_to_data(time, mean_square_displacement)
        else:
            polynomial_coefficients, error_estimate = self._fit_straight_line_to_data(time[fit_range[0]:fit_range[1]], mean_square_displacement[fit_range[0]:fit_range[1]])
        return polynomial_coefficients[0] / 2, error_estimate[0] / 2

    @staticmethod
    def _fit_straight_line_to_data(x, y):
        polynomial_coefficients, covariance_matrix = np.polyfit(x, y, 1, cov=True)
        error_estimate = [np.sqrt(covariance_matrix[0, 0]), np.sqrt(covariance_matrix[1, 1])]
        return polynomial_coefficients, error_estimate

    def calculate_diffusion_coefficient_using_covariance_based_estimator(self, R=None):
        """
        Unbiased estimator of the diffusion coefficient. More info at `https://www.nature.com/articles/nmeth.2904`.
        If the motion blur coefficient is entered a variance estimate is also calculated.

        R: float, motion blur coefficient

        Returns
        -------
            diffusion_coefficient: float
            variance_estimate: float
        """
        squared_displacements = []
        covariance_term = []

        for index, first_position in enumerate(self.particle_positions[:-2]):
            second_position = self.particle_positions[index + 1]
            third_position = self.particle_positions[index + 2]
            if first_position['frame_index'] - second_position['frame_index'] == -1 and first_position['frame_index'] - third_position['frame_index'] == -2:
                squared_displacements.append((self.pixel_width * (second_position['position'] - first_position['position'])) ** 2)
                covariance_term.append((second_position['position'] - first_position['position']) * (
                        third_position['position'] - second_position['position']) * self.pixel_width ** 2)

        time_step = self._calculate_time_step()
        number_of_points_used = len(squared_displacements)
        diffusion_coefficient = np.mean(squared_displacements) / (2 * time_step) + np.mean(covariance_term) / time_step

        if R is not None:
            localisation_error = R * np.mean(squared_displacements) + (2 * R - 1) * np.mean(covariance_term)
            epsilon = localisation_error ** 2 / (diffusion_coefficient * time_step) - 2 * R
            variance_estimate = diffusion_coefficient ** 2 * ((6 + 4 * epsilon + 2 * epsilon ** 2) / number_of_points_used + 4 * (1 + epsilon) ** 2 / (number_of_points_used ** 2))
            return diffusion_coefficient, variance_estimate

        return diffusion_coefficient

    def _calculate_time_step(self):
        return (self.particle_positions['time'][1] - self.particle_positions['time'][0]) / (self.particle_positions['frame_index'][1] - self.particle_positions['frame_index'][0])

    @staticmethod
    def _find_index_of_first_not_equal_element(p1, p2):
        n = 0
        length_of_shortest_trajectory = min(p1.shape[0], p2.shape[0])
        while (n < length_of_shortest_trajectory) and (p1[n] == p2[n]):
            n += 1
        return n

    @staticmethod
    def _find_last_index_where_no_overlaps_occurs(p1, p2):
        n1 = 0
        n2 = 0
        print(p1)
        print(p2)
        while ((n1 < p1.shape[0]) or (n2 < p2.shape[0])) and (p1[n1] != p2[n2]):
            if p1[n1]['frame_index'] < p2[n2]['frame_index']:
                n1 += 1
            elif p1[n1]['frame_index'] > p2[n2]['frame_index']:
                n2 += 1
            else:
                n1 += 1
                n2 += 1
        return n1, n2



import unittest
import numpy as np
from particle_tracker_one_d import ShortestPathFinder
import pprint


class SetAttributeTester(unittest.TestCase):
    mock_frames = np.zeros((2, 3), dtype=np.float32)

    def test_validation_of_frames_argument(self):
        """
        Test that the frames are a numpy array with shape (nFrames,nPixels) with float values between 0 and 1.0.
        The number of frames nFrames should be larger than one and the nPixels should be larger than 2.
        The frames should contain at least one value equal to 1 and one value equal to 0.
        """
        valid_frames = [
            np.array([[0, 1, 0], [0, 0, 0]], dtype=np.float32),
            np.array([[0.1, 0, 0.34], [0.1, 0, 1], [0.12, 0.003, 0.9]], dtype=np.float32),
        ]
        non_valid_shape_or_value_frames = [
            np.array([[0, 1, 0]], dtype=np.float32),
            np.array([0, 1, 0, 0.1], dtype=np.float32),
            np.array([[0.2], [0.4]], dtype=np.float32),
        ]
        non_valid_type_of_frames = [
            2,
            [[0, 1, 0], [0, 0, 0]],
            'string'
        ]
        for frames in valid_frames:
            self.assertTrue(ShortestPathFinder._test_if_frames_have_correct_format(frames), msg='Valid frames not accepted, frames: ' + np.array_str(frames))

        for frames in non_valid_shape_or_value_frames:
            with self.assertRaises(ValueError):
                ShortestPathFinder._test_if_frames_have_correct_format(frames)

        for frames in non_valid_type_of_frames:
            with self.assertRaises(TypeError):
                ShortestPathFinder._test_if_frames_have_correct_format(frames)

    def test_validation_time_argument(self):
        """
        Test that the validation of the class argument works. The time should be a numpy.ndarray with monotonically increasing values.
        """
        valid_times = [
            np.array([0, 1, 2, 4, 6], dtype=np.float32),
            np.array([0, 1000], dtype=np.float32),
            np.array([0.1, 0.2, 0.4], dtype=np.float32),
            np.array([1, 2, 3], dtype=np.float32)
        ]

        non_valid_shape_or_values_times = [
            np.array([0, 0], dtype=np.float32),
            np.array([0, -1], dtype=np.float32),
            np.array([[0, 2], [3, 5]], dtype=np.float32)
        ]

        non_valid_types_of_times = [
            '1,2,3,4',
            [1, 2, 3, 4],
            1,
            {'1': 1, '2': 2}
        ]

        for times in valid_times:
            self.assertTrue(ShortestPathFinder._test_if_time_has_correct_format(times), msg='Valid times not accepted, times: ' + np.array_str(times))

        for times in non_valid_shape_or_values_times:
            with self.assertRaises(ValueError):
                ShortestPathFinder._test_if_time_has_correct_format(times)

        for times in non_valid_types_of_times:
            with self.assertRaises(TypeError):
                ShortestPathFinder._test_if_time_has_correct_format(times)

    def test_validation_of_time_and_frames_having_same_length(self):
        """ Test that the validation of class arguments time and frames should have the same length."""
        valid_times_and_frames = [
            {
                'time': np.array([0, 1], dtype=np.float32),
                'frames': np.array([[0, 1, 0], [0, 0, 0]], dtype=np.float32)
            },
            {
                'time': np.array([0.1, 0.2, 0.4], dtype=np.float32),
                'frames': np.array([[0.1, 0, 0.34], [0.1, 0, 1], [0.12, 0.003, 0.9]], dtype=np.float32)
            }
        ]

        non_valid_times_and_frames = [
            {
                'time': np.array([0, 1], dtype=np.float32),
                'frames': np.array([[0, 1, 0], [0, 0, 0], [0, 0, 0.4]], dtype=np.float32)
            },
            {
                'time': np.array([0.1, 0.2, 0.4], dtype=np.float32),
                'frames': np.array([[0.1, 0, 0.34], [0.1, 0, 1]], dtype=np.float32)
            }
        ]

        for time_and_frames in valid_times_and_frames:
            self.assertTrue(ShortestPathFinder._test_if_time_and_frames_has_same_length(time_and_frames['time'], time_and_frames['frames']))

        for time_and_frames in non_valid_times_and_frames:
            with self.assertRaises(ValueError):
                ShortestPathFinder._test_if_time_and_frames_has_same_length(time_and_frames['time'], time_and_frames['frames'])

    def test_validation_of_automatic_update(self):

        valid_automatic_update_arguments = [True, False]

        non_valid_automatic_update_arguments = [0, 1, 'True', 'False', np.array, []]

        for automatic_update in valid_automatic_update_arguments:
            self.assertTrue(ShortestPathFinder._test_if_automatic_update_has_correct_format(automatic_update))

        for automatic_update in non_valid_automatic_update_arguments:
            with self.assertRaises(ValueError):
                ShortestPathFinder._test_if_automatic_update_has_correct_format(automatic_update)

    def test_validation_of_setting_the_integration_radius_of_intensity_peaks(self):
        """
        Tests the setting of the class attribute integration_radius_of_intensity_peaks. Should be an integer smaller than half of the number of pixels in a frame.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_integration_radius = [0, 1, 2]
        non_valid_type_of_integration_radius = [1.5, '1', [1, 2]]
        non_valid_values_of_integration_radius = [-1, 3, 100]

        pt = ShortestPathFinder(frames=frames, time=time, automatic_update=automatic_update)

        for radius in valid_integration_radius:
            pt.integration_radius_of_intensity_peaks = radius
            self.assertEqual(pt.integration_radius_of_intensity_peaks, radius)

        for radius in non_valid_type_of_integration_radius:
            with self.assertRaises(TypeError, msg=radius):
                pt.integration_radius_of_intensity_peaks = radius

        for radius in non_valid_values_of_integration_radius:
            with self.assertRaises(ValueError, msg=radius):
                pt.integration_radius_of_intensity_peaks = radius

    def test_validation_of_setting_boxcar_width(self):
        """
        Tests the setting of the class attribute boxcar_width. Should be an integer smaller than the number of pixels in a frame.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_boxcar_widths = [0, 1, 2, 3, 4]
        non_valid_type_of_boxcar_widths = [1.5, '1', [1, 2]]
        non_valid_values_of_boxcar_widths = [-1, 5, 100]

        pt = ShortestPathFinder(frames=frames, time=time, automatic_update=automatic_update)

        for width in valid_boxcar_widths:
            pt.boxcar_width = width
            self.assertEqual(pt.boxcar_width, width)

        for width in non_valid_type_of_boxcar_widths:
            with self.assertRaises(TypeError, msg=width):
                pt.boxcar_width = width

        for width in non_valid_values_of_boxcar_widths:
            with self.assertRaises(ValueError, msg=width):
                pt.boxcar_width = width

    def test_validation_of_setting_start_point(self):
        """
        Tests the setting of the class attribute start_point. Should be a tuple, list or np.array
        """

        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_start_points = [
            (0, 0),
            [0, 1],
            np.array([1, 1], dtype=np.int16),
            np.array([1, 3], dtype=np.int32),
        ]
        non_valid_type_of_start_points = [
            1.5,
            '12',
            None,
            (0.3, 2),
            ['', 4],
            np.array([0, 2], dtype=np.float32)
        ]
        non_valid_values_of_start_points = [
            (0, -1),
            (4, 1),
            (0, 10)
        ]

        spf = ShortestPathFinder(frames=frames, time=time, automatic_update=automatic_update)

        for start_point in valid_start_points:
            spf.start_point = start_point
            self.assertEqual(spf.start_point, (start_point[0], start_point[1]))

        for start_point in non_valid_type_of_start_points:
            with self.assertRaises(TypeError, msg=start_point):
                spf.start_point = start_point

        for start_point in non_valid_values_of_start_points:
            with self.assertRaises(ValueError, msg=start_point):
                spf.start_point = start_point

    def test_validation_of_setting_end_point(self):
        """
        Tests the setting of the class attribute end_point. Should be a tuple, list or np.array
        """

        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_end_points = [
            (0, 0),
            [0, 1],
            np.array([1, 1], dtype=np.int16),
            np.array([1, 3], dtype=np.int32),
            np.array([3, 3], dtype=np.int32)
        ]
        non_valid_type_of_end_points = [
            1.5,
            '12',
            None,
            (0.3, 2),
            ['', 4],
            np.array([0, 2], dtype=np.float32)
        ]
        non_valid_values_of_end_points = [
            (0, -1),
            (4, 1),
            (0, 10)
        ]

        spf = ShortestPathFinder(frames=frames, time=time, automatic_update=automatic_update)

        for end_point in valid_end_points:
            spf.end_point = end_point
            self.assertEqual(spf.end_point, (end_point[0], end_point[1]))

        for end_point in non_valid_type_of_end_points:
            with self.assertRaises(TypeError, msg=end_point):
                spf.end_point = end_point

        for end_point in non_valid_values_of_end_points:
            with self.assertRaises(ValueError, msg=end_point):
                spf.end_point = end_point


class FindParticlePositionsTester(unittest.TestCase):

    def test_finding_local_maximas_function(self):
        """
        Tests finding particle positions. All local maximas should be found, that is, the value should be higher than both the neighbouring value to left and right.
        """
        intensity_examples = [
            np.array([1, 0, 0], dtype=np.float32),
            np.array([0, 1, 0], dtype=np.float32),
            np.array([0, 0, 1], dtype=np.float32),
            np.array([0, 0, 0], dtype=np.float32),
            np.array([0, 1, 1], dtype=np.float32),
            np.array([1, 1, 0], dtype=np.float32),
            np.array([1, 1, 1], dtype=np.float32),
            np.array([1, 0, 1], dtype=np.float32),
            np.array([1, 0, 1, 0], dtype=np.float32),
            np.array([1, 1, 0, 0], dtype=np.float32),
            np.array([1, 0, 0, 1], dtype=np.float32),
            np.array([0, 0, 1, 0], dtype=np.float32),
            np.array([0, 0, 0.3, 0], dtype=np.float32),
        ]
        expected_positions = [
            np.array([], dtype=np.int64),
            np.array([1], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([2], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([2], dtype=np.int64),
            np.array([2], dtype=np.int64)
        ]

        frames_examples = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.1, 0.1, 0.2],
            [0.7, 0.2, 0.7]
        ], dtype=np.float32)

        times_examples = np.array([0, 1, 2, 3, 4])

        spf = ShortestPathFinder(frames=frames_examples, time=times_examples, automatic_update=False)
        spf.particle_detection_threshold = 0.1

        for index, intensity in enumerate(intensity_examples):
            np.testing.assert_array_equal(expected_positions[index], spf._find_local_maximas_larger_than_threshold(intensity))

    def test_finding_initial_particle_positions(self):
        """
        Test finding the initial particle positions, that is local maximas between start point and end point
        """

        frames = np.array([
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
            [0, 1, 0, 1, 0],
            [1, 0, 0, 1, 1],
        ], dtype=np.float32)

        time = np.arange(frames.shape[0])
        automatic_update = False

        start_point = (0, 1)
        end_point = (5, 3)

        expected_positions = [
            np.array([1], dtype=np.float32),
            np.array([1], dtype=np.float32),
            np.array([2], dtype=np.float32),
            np.array([3], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([3], dtype=np.float32),
        ]

        spf = ShortestPathFinder(frames, time, automatic_update)
        spf.start_point = start_point
        spf.end_point = end_point

        spf._find_initial_particle_positions()

        for index, positions in enumerate(spf._particle_positions):
            np.testing.assert_array_equal(expected_positions[index], positions)

    def test_the_refining_of_particle_positions(self):
        """
        Test that the refinement of particle positions work. That is, by finding the center of mass close to the initial particle position.
        """
        automatic_update = False

        frames = np.array([
            [0, 0.1, 0.5, 0],
            [0, 0.6, 0.2, 0],
            [1, 0.1, 0.1, 0],
            [0.1, 0.1, 0.2, 0],
            [0, 0.7, 0.2, 0.7]
        ], dtype=np.float32)

        time = np.array([0, 1, 2, 3, 4])

        expected_positions = [
            np.array([1.8333333], dtype=np.float32),
            np.array([1.25], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([1.6666666], dtype=np.float32),
            np.array([2], dtype=np.float32)
        ]

        spf = ShortestPathFinder(time=time, frames=frames, automatic_update=automatic_update)
        spf.start_point = (0, 2)
        spf.end_point = (4, 2)
        spf.integration_radius_of_intensity_peaks = 1
        spf._find_initial_particle_positions()
        spf._refine_particle_positions()

        self.assertEqual(len(spf.particle_positions), len(expected_positions))

        for index, position in enumerate(spf.particle_positions):
            np.testing.assert_array_equal(expected_positions[index], position)


class AssociationAndCostMatrixTester(unittest.TestCase):

    def test_the_initialisation_of_the_association_and_cost_matrix(self):
        """
        Test that the initialised association matrix has the correct shape.
        """
        automatic_update = False

        frames = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.1, 0.1, 0.2],
            [0.1, 0.1, 0.2],
        ], dtype=np.float32)

        times = np.array([0, 1, 2, 3, 4])

        particle_positions = [
            np.array([0], dtype=np.float32),
            np.array([1], dtype=np.float32),
            np.array([0, 2], dtype=np.float32),
            np.array([0, 2, 3], dtype=np.float32),
            np.array([4], dtype=np.float32),
        ]

        expected_association_matrix = [
            np.zeros((1, 1), dtype=bool),
            np.zeros((1, 2), dtype=bool),
            np.zeros((2, 3), dtype=bool),
            np.zeros((3, 1), dtype=bool)
        ]

        expected_cost_matrix = [
            np.zeros((1, 1), dtype=np.float32),
            np.zeros((1, 2), dtype=np.float32),
            np.zeros((2, 3), dtype=np.float32),
            np.zeros((3, 1), dtype=np.float32)
        ]

        spf = ShortestPathFinder(time=times, frames=frames, automatic_update=automatic_update)
        spf._particle_positions = particle_positions
        spf._initialise_association_and_cost_matrix()

        association_matrix, cost_matrix = spf._association_matrix, spf._cost_matrix

        self.assertEqual(len(expected_association_matrix), len(association_matrix))
        self.assertEqual(len(expected_cost_matrix), len(cost_matrix))

        for frame_index in range(len(association_matrix)):
            np.testing.assert_array_equal(association_matrix[frame_index], expected_association_matrix[frame_index])
            np.testing.assert_array_equal(cost_matrix[frame_index], expected_cost_matrix[frame_index])

    def test_calculation_of_the_cost_matrix(self):
        """
        Test to verify that the calculation of the cost matrix is done correctly.
        """
        frames = np.array([
            [0, 0.3, 0.5],
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.7, 0.2, 0.7],
            [0.1, 0.1, 0.2],
            [0.1, 0.1, 0.2]
        ], dtype=np.float32)

        times = np.array([0, 1, 2, 3, 4, 5, 6])

        particle_positions = [
            np.array([2], dtype=np.float32),
            np.array([1.25], dtype=np.float32),
            np.array([0], dtype=np.float32),
            np.array([0, 2], dtype=np.float32),
            np.array([2], dtype=np.float32)
        ]

        start_point = (1, 2)
        end_point = (5, 2)

        expected_cost_matrix = [
            np.array(
                [
                    [0.06625],
                ], dtype=np.float32),
            np.array(
                [
                    [0.4725],
                ], dtype=np.float32),
            np.array(
                [
                    [0.6325, 1.0325],
                ], dtype=np.float32),
            np.array(
                [
                    [1.1225],
                    [1.1225001],
                ], dtype=np.float32),
        ]

        spf = ShortestPathFinder(frames=frames, time=times, automatic_update=False)
        spf.start_point = start_point
        spf.end_point = end_point

        spf._particle_positions = particle_positions

        spf._initialise_association_and_cost_matrix()
        spf._calculate_particle_moments()
        spf._calculate_cost_matrix()

        print(expected_cost_matrix)
        print(spf._cost_matrix)

        for frame_index, _ in enumerate(spf._cost_matrix):
            for future_frame_index, _ in enumerate(spf._cost_matrix[frame_index]):
                np.testing.assert_array_almost_equal(spf._cost_matrix[frame_index][future_frame_index], expected_cost_matrix[frame_index][future_frame_index])

    def test_find_shortest_path_by_dijkstra(self):
        """
         Test automatic finding
         """

        automatic_update = True

        frames = np.array([
            [0, 1, 0, 0, 1, 0],
            [0, 0, 1, 0, 1, 0],
            [0, 1, 0, 1, 0, 0],
            [0, 0, 1, 0, 1, 0],
            [0, 0, 0, 0, 1, 0]
        ], dtype=np.float32)

        times = np.array([0, 1, 2, 3, 4])

        start_point = (0, 1)
        end_point = (4, 4)

        spf = ShortestPathFinder(frames=frames, time=times, automatic_update=automatic_update)
        spf.boxcar_width = 0
        spf.start_point = start_point
        spf.end_point = end_point

        #print('1.')
        pprint.pprint(spf._create_trajectory_from_cost_matrix())
        #print('2.')



import unittest
import numpy as np
from particle_tracker_one_d import Trajectory


class AddTrajectoriesTester(unittest.TestCase):

    def test_addition_with_empty_trajectory(self):
        """
        Test adding one empty trajectory with a non empty trajectory
        """

        particle_positions = np.empty((1,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        particle_positions['frame_index'] = np.array([0])
        particle_positions['time'] = np.array([1])
        particle_positions['position'] = np.array([0])

        empty_trajectory = Trajectory(pixel_width=2)
        non_empty_trajectory = Trajectory(pixel_width=2)
        non_empty_trajectory.particle_positions = particle_positions

        final_trajectory = empty_trajectory + non_empty_trajectory

        np.testing.assert_array_equal(non_empty_trajectory.particle_positions, (non_empty_trajectory + empty_trajectory)._particle_positions)
        np.testing.assert_array_equal(non_empty_trajectory.particle_positions, (empty_trajectory + non_empty_trajectory)._particle_positions)
        np.testing.assert_array_equal(empty_trajectory.particle_positions, (empty_trajectory + empty_trajectory)._particle_positions)

    def test_addition_with_non_empty_trajectories(self):
        """
        Test adding together two trajectories
        """

        particle_positions_1 = np.empty((1,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        particle_positions_1['frame_index'] = np.array([0])
        particle_positions_1['time'] = np.array([1])
        particle_positions_1['position'] = np.array([0])

        particle_positions_2 = np.empty((1,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        particle_positions_2['frame_index'] = np.array([1])
        particle_positions_2['time'] = np.array([2])
        particle_positions_2['position'] = np.array([1])

        final_positions = np.empty((2,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        final_positions['frame_index'] = np.array([0, 1])
        final_positions['time'] = np.array([1, 2])
        final_positions['position'] = np.array([0, 1])

        t_1 = Trajectory()
        t_1.particle_positions = particle_positions_1
        t_2 = Trajectory()
        t_2.particle_positions = particle_positions_2

        np.testing.assert_array_equal(final_positions, (t_1 + t_2)._particle_positions)
        np.testing.assert_array_equal(final_positions, (t_2 + t_1)._particle_positions)

    def test_adding_overlapping_trajectories(self):
        """
        Test addition of two overlapping trajectories
        """

        particle_positions_1 = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        particle_positions_1['frame_index'] = np.array([0, 1, 2])
        particle_positions_1['time'] = np.array([1, 2, 3])
        particle_positions_1['position'] = np.array([0, 3, 1])

        t_1 = Trajectory()
        t_1.particle_positions = particle_positions_1

        particle_positions_2 = np.empty((3,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        particle_positions_2['frame_index'] = np.array([1, 2, 3])
        particle_positions_2['time'] = np.array([2, 3, 4])
        particle_positions_2['position'] = np.array([5, 2, 0])

        t_2 = Trajectory()
        t_2.particle_positions = particle_positions_2

        final_positions = np.empty((4,), dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32)])

        final_positions['frame_index'] = np.array([0, 1, 2, 3])
        final_positions['time'] = np.array([1, 2, 3, 4])
        final_positions['position'] = np.array([0, 5, 2, 0])


        np.testing.assert_array_equal(final_positions, (t_1 + t_2)._particle_positions)
        np.testing.assert_array_equal(final_positions, (t_2 + t_1)._particle_positions)



import numpy as np


class Frames:
    """
    Class for handling frames, with averaging and easy plot functions.

    Parameters
    ----------
    frames: np.array
        The frames in which trajectories are to be found. The shape of the np.array should be (nFrames,xPixels). The intensity of the frames should be normalised according to
        :math:`I_n = (I-I_{min})/(I_{max}-I_{min})`, where :math:`I` is the intensity of the frames, :math:`I_{min}`, :math:`I_{max}` are the global intensity minima and maxima of the
        frames.
    time: np.array
        The corresponding time of each frame.
    automatic_update: bool
        Should the class update automatically when the property boxcar_width is changed. Default=True

    Attributes
    ----------
    frames
    time
    boxcar_width
    """

    def __init__(self, frames, time, automatic_update=True):
        Frames._validate_class_arguments(frames, time, automatic_update)
        self._automatic_update = automatic_update
        self._frames = frames
        self._averaged_intensity = frames
        self.time = time
        self._boxcar_width = 0

    @property
    def frames(self):
        """
        np.array:
            The frames which have been restored by boxcar averaging.
        """
        return self._averaged_intensity

    @property
    def boxcar_width(self):
        """
        int:
            Number of values used in the boxcar averaging of the frames.
        """
        return self._boxcar_width

    @boxcar_width.setter
    def boxcar_width(self, width):
        if type(width) is not int:
            raise TypeError('Attribute boxcar_width should be of type int')
        if not -1 < width <= self.frames.shape[1]:
            raise ValueError('Attribute boxcar_width should be a positive integer less or equal the number of pixels in each frame.')

        if not width == self._boxcar_width:
            self._boxcar_width = width
            if self._automatic_update:
                self._update_averaged_intensity()

    def plot_all_frames(self, ax=None, **kwargs):
        """
        ax: matplotlib axes instance
            The axes which you want the frames to plotted on. If none is provided a new instance will be created.
        **kwargs:
            Plot settings, any settings which can be used in matplotlib.pyplot.imshow method.

        Returns
        -------
            matplotlib axes instance
                Returns the axes input argument or creates and returns a new instance of an matplotlib axes object.
        """
        if ax is None:
            ax = plt.axes()
        ax.imshow(self._averaged_intensity, **kwargs)
        return ax

    def _update_averaged_intensity(self):
        if self.boxcar_width == 0 or self.boxcar_width == 1:
            self._averaged_intensity = self._frames
        else:
            self._averaged_intensity = np.empty((self._frames.shape[0],self._frames.shape[1]-self._boxcar_width+1), dtype=np.float32)
            kernel = np.ones((self.boxcar_width,)) / self.boxcar_width
            for row_index, row_intensity in enumerate(self._frames):
                self._averaged_intensity[row_index] = np.convolve(row_intensity, kernel, mode='valid')

    @staticmethod
    def _validate_class_arguments(frames, time, automatic_update):
        Frames._test_if_frames_have_correct_format(frames)
        Frames._test_if_time_has_correct_format(time)
        Frames._test_if_time_and_frames_has_same_length(time, frames)
        Frames._test_if_automatic_update_has_correct_format(automatic_update)

    @staticmethod
    def _test_if_frames_have_correct_format(frames):
        if type(frames) is not np.ndarray:
            raise TypeError('Class argument frames not of type np.ndarray')
        if not (len(frames.shape) == 2 and frames.shape[0] >= 1 and frames.shape[1] >= 1):
            raise ValueError(
                'Class argument frames need to be of shape (nFrames,nPixels) with nFrames >= 1 and nPixels >= 1')
        return True

    @staticmethod
    def _test_if_time_has_correct_format(time):
        if type(time) is not np.ndarray:
            raise TypeError('Class argument frames not of type np.ndarray')
        if not (len(time.shape) == 1 and time.shape[0] >= 1):
            raise ValueError('Class argument time need to be of shape (nFrames,) with nFrames >= 1.')
        if not all(np.diff(time) > 0) and not time.shape[0] == 1:
            raise ValueError('Class argument time not increasing monotonically.')
        return True

    @staticmethod
    def _test_if_time_and_frames_has_same_length(time, frames):
        if not time.shape[0] == frames.shape[0]:
            raise ValueError('Class arguments time and frames does not of equal length.')
        return True

    @staticmethod
    def _test_if_automatic_update_has_correct_format(automatic_update):
        if not type(automatic_update) == bool:
            raise ValueError('Class argument automatic_update must be True or False.')
        return True



import unittest
import numpy as np
from particle_tracker_one_d import ParticleTracker


class SetAttributeTester(unittest.TestCase):
    mock_frames = np.zeros((2, 3), dtype=np.float32)

    def test_validation_of_frames_argument(self):
        """
        Test that the frames are a numpy array with shape (nFrames,nPixels) with float values between 0 and 1.0.
        The number of frames nFrames should be larger than one and the nPixels should be larger than 2.
        """
        valid_frames = [
            np.array([[0, 1, 0], [0, 0, 0]], dtype=np.float32),
            np.array([[0.1, 0, 0.34], [0.1, 0, 1], [0.12, 0.003, 0.9]], dtype=np.float32),
        ]
        non_valid_shape_or_value_frames = [
            np.array([[0, 1, 0]], dtype=np.float32),
            np.array([0, 1, 0, 0.1], dtype=np.float32),
            np.array([[0.2], [0.4]], dtype=np.float32),
        ]
        non_valid_type_of_frames = [
            2,
            [[0, 1, 0], [0, 0, 0]],
            'string'
        ]
        for frames in valid_frames:
            self.assertTrue(ParticleTracker._test_if_frames_have_correct_format(frames), msg='Valid frames not accepted, frames: ' + np.array_str(frames))

        for frames in non_valid_shape_or_value_frames:
            with self.assertRaises(ValueError):
                ParticleTracker._test_if_frames_have_correct_format(frames)

        for frames in non_valid_type_of_frames:
            with self.assertRaises(TypeError):
                ParticleTracker._test_if_frames_have_correct_format(frames)

    def test_validation_time_argument(self):
        """
        Test that the validation of the class argument works. The time should be a numpy.ndarray with monotonically increasing values.
        """
        valid_times = [
            np.array([0, 1, 2, 4, 6], dtype=np.float32),
            np.array([0, 1000], dtype=np.float32),
            np.array([0.1, 0.2, 0.4], dtype=np.float32),
            np.array([1, 2, 3], dtype=np.float32)
        ]

        non_valid_shape_or_values_times = [
            np.array([0, 0], dtype=np.float32),
            np.array([0, -1], dtype=np.float32),
            np.array([[0, 2], [3, 5]], dtype=np.float32)
        ]

        non_valid_types_of_times = [
            '1,2,3,4',
            [1, 2, 3, 4],
            1,
            {'1': 1, '2': 2}
        ]

        for times in valid_times:
            self.assertTrue(ParticleTracker._test_if_time_has_correct_format(times), msg='Valid times not accepted, times: ' + np.array_str(times))

        for times in non_valid_shape_or_values_times:
            with self.assertRaises(ValueError):
                ParticleTracker._test_if_time_has_correct_format(times)

        for times in non_valid_types_of_times:
            with self.assertRaises(TypeError):
                ParticleTracker._test_if_time_has_correct_format(times)

    def test_validation_of_time_and_frames_having_same_length(self):
        """ Test that the validation of class arguments time and frames should have the same length."""
        valid_times_and_frames = [
            {
                'time': np.array([0, 1], dtype=np.float32),
                'frames': np.array([[0, 1, 0], [0, 0, 0]], dtype=np.float32)
            },
            {
                'time': np.array([0.1, 0.2, 0.4], dtype=np.float32),
                'frames': np.array([[0.1, 0, 0.34], [0.1, 0, 1], [0.12, 0.003, 0.9]], dtype=np.float32)
            }
        ]

        non_valid_times_and_frames = [
            {
                'time': np.array([0, 1], dtype=np.float32),
                'frames': np.array([[0, 1, 0], [0, 0, 0], [0, 0, 0.4]], dtype=np.float32)
            },
            {
                'time': np.array([0.1, 0.2, 0.4], dtype=np.float32),
                'frames': np.array([[0.1, 0, 0.34], [0.1, 0, 1]], dtype=np.float32)
            }
        ]

        for time_and_frames in valid_times_and_frames:
            self.assertTrue(ParticleTracker._test_if_time_and_frames_has_same_length(time_and_frames['time'], time_and_frames['frames']))

        for time_and_frames in non_valid_times_and_frames:
            with self.assertRaises(ValueError):
                ParticleTracker._test_if_time_and_frames_has_same_length(time_and_frames['time'], time_and_frames['frames'])

    def test_validation_of_automatic_update(self):

        valid_automatic_update_arguments = [True, False]

        non_valid_automatic_update_arguments = [0, 1, 'True', 'False', np.array, []]

        for automatic_update in valid_automatic_update_arguments:
            self.assertTrue(ParticleTracker._test_if_automatic_update_has_correct_format(automatic_update))

        for automatic_update in non_valid_automatic_update_arguments:
            with self.assertRaises(ValueError):
                ParticleTracker._test_if_automatic_update_has_correct_format(automatic_update)

    def test_validation_of_setting_the_integration_radius_of_intensity_peaks(self):
        """
        Tests the setting of the class attribute integration_radius_of_intensity_peaks. Should be an integer smaller than half of the number of pixels in a frame.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_integration_radius = [0, 1, 2]
        non_valid_type_of_integration_radius = [1.5, '1', [1, 2]]
        non_valid_values_of_integration_radius = [-1, 3, 100]

        pt = ParticleTracker(frames=frames, time=time, automatic_update=automatic_update)

        for radius in valid_integration_radius:
            pt.integration_radius_of_intensity_peaks = radius
            self.assertEqual(pt.integration_radius_of_intensity_peaks, radius)

        for radius in non_valid_type_of_integration_radius:
            with self.assertRaises(TypeError, msg=radius):
                pt.integration_radius_of_intensity_peaks = radius

        for radius in non_valid_values_of_integration_radius:
            with self.assertRaises(ValueError, msg=radius):
                pt.integration_radius_of_intensity_peaks = radius

    def test_validation_of_setting_boxcar_width(self):
        """
        Tests the setting of the class attribute boxcar_width. Should be an integer smaller than the number of pixels in a frame.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_boxcar_widths = [0, 1, 2, 3, 4]
        non_valid_type_of_boxcar_widths = [1.5, '1', [1, 2]]
        non_valid_values_of_boxcar_widths = [-1, 5, 100]

        pt = ParticleTracker(frames=frames, time=time, automatic_update=automatic_update)

        for width in valid_boxcar_widths:
            pt.boxcar_width = width
            self.assertEqual(pt.boxcar_width, width)

        for width in non_valid_type_of_boxcar_widths:
            with self.assertRaises(TypeError, msg=width):
                pt.boxcar_width = width

        for width in non_valid_values_of_boxcar_widths:
            with self.assertRaises(ValueError, msg=width):
                pt.boxcar_width = width

    def test_validation_of_setting_particle_detection_threshold(self):
        """
        Tests the setting of the class attribute particle_detection_threshold. Should be a numerical value between 0 and 1.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_particle_detection_thresholds = [0, 0.1, 0.22, 0.93, 1]
        non_valid_type_detection_thresholds = ['1', [1, 2], None]
        non_valid_values_of_detection_thresholds = [-1, 5, 100]

        pt = ParticleTracker(frames=frames, time=time, automatic_update=automatic_update)

        for threshold in valid_particle_detection_thresholds:
            pt.particle_detection_threshold = threshold
            self.assertEqual(pt.particle_detection_threshold, threshold)

        for threshold in non_valid_type_detection_thresholds:
            with self.assertRaises(TypeError, msg=threshold):
                pt.particle_detection_threshold = threshold

        for threshold in non_valid_values_of_detection_thresholds:
            with self.assertRaises(ValueError, msg=threshold):
                pt.particle_detection_threshold = threshold

    def test_validation_of_setting_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles(self):
        """
        Tests the setting of the class attribute maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles. Should be a numerical value between 0 and 1.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = [0, 1, 2, 3]
        non_valid_type_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = ['1', [1, 2], None, 1.2]
        non_valid_values_of_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = [-1, 5, 100]

        pt = ParticleTracker(frames=frames, time=time, automatic_update=automatic_update)

        for number in valid_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles:
            pt.maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = number
            self.assertEqual(pt.maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles, number)

        for number in non_valid_type_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles:
            with self.assertRaises(TypeError, msg=number):
                pt.maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = number

        for number in non_valid_values_of_maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles:
            with self.assertRaises(ValueError, msg=number):
                pt.maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = number

    def test_validation_of_setting_maximum_distance_a_particle_can_travel_between_frames(self):
        """
        Tests the setting of the class attribute maximum_distance_a_particle_can_travel_between_frames. Should be a numerical value between 0 and number of pixels in each frame.
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        valid_maximum_distance_a_particle_can_travel_between_frames = [0.4, 1.4, 2, 3]
        non_valid_type_maximum_distance_a_particle_can_travel_between_frames = ['1', [1, 2], None]
        non_valid_values_of_maximum_distance_a_particle_can_travel_between_frames = [-1, 5, 100]

        pt = ParticleTracker(frames=frames, time=time, automatic_update=automatic_update)

        for distance in valid_maximum_distance_a_particle_can_travel_between_frames:
            pt.maximum_distance_a_particle_can_travel_between_frames = distance
            self.assertEqual(pt.maximum_distance_a_particle_can_travel_between_frames, distance)

        for distance in non_valid_type_maximum_distance_a_particle_can_travel_between_frames:
            with self.assertRaises(TypeError, msg=distance):
                pt.maximum_distance_a_particle_can_travel_between_frames = distance

        for distance in non_valid_values_of_maximum_distance_a_particle_can_travel_between_frames:
            with self.assertRaises(ValueError, msg=distance):
                pt.maximum_distance_a_particle_can_travel_between_frames = distance

    def test_changing_the_cost_coefficients(self):
        """
        Test that it is possible to change cost coefficients
        """
        frames = np.array([
            [0, 0.1, 0.2, 0.1],
            [0, 0.2, 0.3, 0.4],
            [0.2, 0.5, 0.6, 1],
            [0, 0.1, 0.2, 0.1]
        ], dtype=np.float32)
        time = np.array([0, 1, 2, 3])
        automatic_update = False

        correct_cost_coefficients = np.array([
            [1, 2, 3],
            [0, 0.2, 0.1],
            [10, 10.2, -1]
        ])

        incorrect_cost_coefficients = np.array([
            [0, 0, 0]
        ])

        pt = ParticleTracker(frames=frames, time=time, automatic_update=automatic_update)

        for cost_coefficients in correct_cost_coefficients:
            pt.change_cost_coefficients(cost_coefficients[0], cost_coefficients[1], cost_coefficients[2])
            np.testing.assert_array_almost_equal(cost_coefficients, pt._cost_coefficients)

        for cost_coefficients in incorrect_cost_coefficients:
            with self.assertRaises(ValueError):
                pt.change_cost_coefficients(cost_coefficients[0], cost_coefficients[1], cost_coefficients[2])


class FindParticlePositionsTester(unittest.TestCase):

    def test_finding_local_maximas_function(self):
        """
        Tests finding particle positions. All local maximas should be found, that is, the value should be higher than both the neighbouring value to left and right.
        """
        intensity_examples = [
            np.array([1, 0, 0], dtype=np.float32),
            np.array([0, 1, 0], dtype=np.float32),
            np.array([0, 0, 1], dtype=np.float32),
            np.array([0, 0, 0], dtype=np.float32),
            np.array([0, 1, 1], dtype=np.float32),
            np.array([1, 1, 0], dtype=np.float32),
            np.array([1, 1, 1], dtype=np.float32),
            np.array([1, 0, 1], dtype=np.float32),
            np.array([1, 0, 1, 0], dtype=np.float32),
            np.array([1, 1, 0, 0], dtype=np.float32),
            np.array([1, 0, 0, 1], dtype=np.float32),
            np.array([0, 0, 1, 0], dtype=np.float32),
            np.array([0, 0, 0.3, 0], dtype=np.float32),
        ]
        expected_positions = [
            np.array([], dtype=np.int64),
            np.array([1], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([2], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([2], dtype=np.int64),
            np.array([2], dtype=np.int64)
        ]

        frames_examples = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.1, 0.1, 0.2],
            [0.7, 0.2, 0.7]
        ], dtype=np.float32)

        times_examples = np.array([0, 1, 2, 3, 4])

        pt = ParticleTracker(frames=frames_examples, time=times_examples, automatic_update=False)
        pt.particle_detection_threshold = 0.1

        for index, intensity in enumerate(intensity_examples):
            np.testing.assert_array_equal(expected_positions[index], pt._find_local_maximas_larger_than_threshold(intensity))

    def test_finding_initial_particle_positions_function(self):
        """
        Tests finding the initial particle positions before any discrimination and refinement of positions.
        """
        frames_examples = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.1, 0.1, 0.2],
            [0.7, 0.2, 0.7]
        ], dtype=np.float32)

        times_examples = np.array([0, 1, 2, 3, 4])

        expected_positions = [
            np.array([], dtype=np.float32),
            np.array([1], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([], dtype=np.float32)
        ]

        automatic_update = False

        pt = ParticleTracker(time=times_examples, frames=frames_examples, automatic_update=automatic_update)
        pt.particle_detection_threshold = 0.3
        pt._find_initial_particle_positions()

        self.assertEqual(len(pt.particle_positions), len(expected_positions))

        for index, position in enumerate(pt.particle_positions):
            np.testing.assert_array_equal(expected_positions[index], position)

    def test_find_center_of_mass_function(self):
        """
        Test finding the center of mass close to the initial particle positions.
        """

        intensity_examples = [
            np.array([1, 0, 0], dtype=np.float32),
            np.array([0, 1, 0], dtype=np.float32),
            np.array([0, 0, 1], dtype=np.float32),
            np.array([1, 1, 0], dtype=np.float32)
        ]

        expected_center_of_mass = np.array([0, 1, 2, 0.5], dtype=np.float32)

        for index, intensity in enumerate(intensity_examples):
            self.assertEqual(ParticleTracker._calculate_center_of_mass(intensity), expected_center_of_mass[index])

    def test_the_refining_of_particle_positions(self):
        """
        Test that the refinement of particle positions work. That is, by finding the center of mass close to the initial particle position.
        """
        automatic_update = False

        frames_example = np.array([
            [0, 0.1, 0.5, 0],
            [0, 0.6, 0.2, 0],
            [1, 0.1, 0.1, 0],
            [0.1, 0.1, 0.2, 0],
            [0, 0.7, 0.2, 0.7]
        ], dtype=np.float32)

        times_example = np.array([0, 1, 2, 3, 4])

        expected_positions = [
            np.array([1.8333333], dtype=np.float32),
            np.array([1.25], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([1.2222222], dtype=np.float32)
        ]

        pt = ParticleTracker(time=times_example, frames=frames_example, automatic_update=automatic_update)
        pt.particle_detection_threshold = 0.3
        pt.integration_radius_of_intensity_peaks = 1
        pt._find_initial_particle_positions()
        pt._refine_particle_positions()

        self.assertEqual(len(pt.particle_positions), len(expected_positions))

        for index, position in enumerate(pt.particle_positions):
            np.testing.assert_array_equal(expected_positions[index], position)


class AssociationAndCostMatrixTester(unittest.TestCase):

    def test_the_initialisation_of_the_association_and_cost_matrix(self):
        """
        Test that the initialised association matrix has the correct shape.
        """
        automatic_update = False

        frames_example = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.1, 0.1, 0.2],
        ], dtype=np.float32)

        times_example = np.array([0, 1, 2, 3])

        particle_positions = [
            np.array([], dtype=np.float32),
            np.array([1], dtype=np.float32),
            np.array([0, 2], dtype=np.float32),
            np.array([0, 2, 3], dtype=np.float32),
        ]

        expected_association_matrix = [
            [
                np.zeros((1, 2), dtype=bool),
                np.zeros((1, 3), dtype=bool),
                np.zeros((1, 4), dtype=bool)
            ],
            [
                np.zeros((2, 3), dtype=bool),
                np.zeros((2, 4), dtype=bool)
            ],
            [
                np.zeros((3, 4), dtype=bool)
            ],
            []
        ]

        expected_cost_matrix = [
            [
                np.zeros((1, 2), dtype=np.float32),
                np.zeros((1, 3), dtype=np.float32),
                np.zeros((1, 4), dtype=np.float32)
            ],
            [
                np.zeros((2, 3), dtype=np.float32),
                np.zeros((2, 4), dtype=np.float32)
            ],
            [
                np.zeros((3, 4), dtype=np.float32)
            ],
            []
        ]

        pt = ParticleTracker(time=times_example, frames=frames_example, automatic_update=automatic_update)
        pt.maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = 2
        pt._particle_positions = particle_positions

        pt._initialise_association_and_cost_matrices()

        self.assertEqual(len(expected_association_matrix), len(pt._association_matrix))
        self.assertEqual(len(expected_cost_matrix), len(pt._cost_matrix))

        for index in range(len(expected_association_matrix)):
            self.assertEqual(len(expected_association_matrix[index]), len(pt._association_matrix[index]), msg=index)
            self.assertEqual(len(expected_cost_matrix[index]), len(pt._cost_matrix[index]), msg=index)

        for frame_index in range(len(expected_association_matrix)):
            for r in range(len(expected_association_matrix[frame_index])):
                np.testing.assert_array_equal(pt._association_matrix[frame_index][r], expected_association_matrix[frame_index][r])
                np.testing.assert_array_equal(pt._cost_matrix[frame_index][r], expected_cost_matrix[frame_index][r])

    def test_calculation_of_zeroth_order_intensity_moments(self):
        """
        Verification of the calulation of first order intensity moments
        TODO: Rewrite to have more logical frames with particle positions going from left to right.
        """

        frames_example = np.array([
            [0.3, 0.1, 0, 0.1, 0.1, 0.5],
            [0, 0.6, 0.2, 0.3, 0.1, 0],
            [1, 0.1, 0.1, 0.3, 0.2, 0],
            [0.1, 0.1, 0.2, 0.2, 0.1, 0.1],
            [0.1, 0.1, 0.8, 0.2, 0.3, 0],
            [0.7, 0.2, 0.2, 0.2, 0.6, 0.1],
            [0.1, 0.8, 0.2, 0.2, 0.9, 0.1]
        ], dtype=np.float32)

        times_example = np.array([0, 1, 2, 3, 4, 5, 6])

        particle_positions = np.array([
            [5, 0],
            [1.25, 1],
            [0, 2],
            [2, 4],
            [0, 5],
            [4, 5],
            [1, 6],
            [4, 6]
        ], dtype=np.float32)

        expected_zeroth_order_intensity_moment_integration_radius_zero = [
            0.5,
            0.6,
            1,
            0.8,
            0.7,
            0.6,
            0.8,
            0.9
        ]

        expected_zeroth_order_intensity_moment_integration_radius_one = [
            0.7,
            0.8,
            1.2,
            1.1,
            1.1,
            0.9,
            1.1,
            1.2
        ]

        expected_zeroth_order_intensity_moment_integration_radius_two = [
            0.9,
            1.4,
            1.4,
            1.5,
            1.5,
            1.3,
            1.5,
            1.6
        ]

        expected_zeroth_order_intensity_moment_integration_radius_three = [
            3.3,
            3.2
        ]

        expected_second_order_intensity_moment_integration_radius_zero = [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0
        ]

        expected_second_order_intensity_moment_integration_radius_one = [
            0.2857142857,
            0.25,
            0.1666666667,
            0.2727272727,
            0.3636363636,
            0.3333333333,
            0.2727272727,
            0.25
        ]

        expected_second_order_intensity_moment_integration_radius_two = [
            1.1111111111,
            1.8571428571,
            0.7142857143,
            1.2666666667,
            1.3333333333,
            1.4615384615,
            1.2666666667,
            1.1875
        ]

        expected_second_order_intensity_moment_integration_radius_three = [
            5.48484848485,
            5.09375
        ]

        pt = ParticleTracker(frames=frames_example, time=times_example, automatic_update=False)
        pt.integration_radius_of_intensity_peaks = 0

        for index, position in enumerate(particle_positions):
            np.testing.assert_almost_equal(pt._calculate_zeroth_order_intensity_moment(position[0], int(round(position[1]))),
                                           expected_zeroth_order_intensity_moment_integration_radius_zero[index])

        for index, position in enumerate(particle_positions):
            np.testing.assert_almost_equal(pt._calculate_second_order_intensity_moment(position[0], int(round(position[1]))),
                                           expected_second_order_intensity_moment_integration_radius_zero[index])

        pt.integration_radius_of_intensity_peaks = 1

        for index, position in enumerate(particle_positions):
            np.testing.assert_almost_equal(expected_zeroth_order_intensity_moment_integration_radius_one[index],
                                           pt._calculate_zeroth_order_intensity_moment(position[0], int(round(position[1]))))

        for index, position in enumerate(particle_positions):
            np.testing.assert_almost_equal(pt._calculate_second_order_intensity_moment(position[0], int(round(position[1]))),
                                           expected_second_order_intensity_moment_integration_radius_one[index])

        pt.integration_radius_of_intensity_peaks = 2

        for index, position in enumerate(particle_positions):
            np.testing.assert_almost_equal(pt._calculate_zeroth_order_intensity_moment(position[0], int(round(position[1]))),
                                           expected_zeroth_order_intensity_moment_integration_radius_two[index], decimal=5)

        for index, position in enumerate(particle_positions):
            np.testing.assert_almost_equal(pt._calculate_second_order_intensity_moment(position[0], int(round(position[1]))),
                                           expected_second_order_intensity_moment_integration_radius_two[index])

        pt.integration_radius_of_intensity_peaks = 3

        np.testing.assert_almost_equal(pt._calculate_zeroth_order_intensity_moment(1, 6),
                                       expected_zeroth_order_intensity_moment_integration_radius_three[0])

        np.testing.assert_almost_equal(pt._calculate_zeroth_order_intensity_moment(4, 6),
                                       expected_zeroth_order_intensity_moment_integration_radius_three[1])

        np.testing.assert_almost_equal(pt._calculate_second_order_intensity_moment(1, 6),
                                       expected_second_order_intensity_moment_integration_radius_three[0], decimal=5)

        np.testing.assert_almost_equal(pt._calculate_second_order_intensity_moment(4, 6),
                                       expected_second_order_intensity_moment_integration_radius_three[1])

    def test_calculation_of_the_cost_matrix(self):
        """
        Test to verify that the calculation of the cost matrix is done correctly.
        """
        frames_example = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [0.1, 0.1, 0.2],
            [0.7, 0.2, 0.7]
        ], dtype=np.float32)

        times_example = np.array([0, 1, 2, 3, 4])

        particle_positions = [
            np.array([2], dtype=np.float32),
            np.array([1.25], dtype=np.float32),
            np.array([0], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([0, 2], dtype=np.float32)
        ]

        max_frames = 2
        max_distance = 1

        expected_cost_matrix = [
            [
                np.array(
                    [
                        [0, 1],
                        [1, 0.5737755],
                    ], dtype=np.float32),
                np.array(
                    [
                        [0, 4],
                        [4, np.inf]
                    ], dtype=np.float32),
                np.array(
                    [
                        [0],
                        [9]
                    ], dtype=np.float32)
            ],
            [
                np.array(
                    [
                        [0, 1],
                        [1, np.inf]
                    ], dtype=np.float32),
                np.array(
                    [
                        [0],
                        [4]
                    ], dtype=np.float32),
                np.array(
                    [
                        [0, 9, 9],
                        [9, 1.6654132, 0.6654132]
                    ], dtype=np.float32)

            ],
            [
                np.array(
                    [
                        [0],
                        [1]
                    ], dtype=np.float32),
                np.array(
                    [
                        [0, 4, 4],
                        [4, 0.0487971, np.inf]
                    ], dtype=np.float32)

            ],
            [
                np.array(
                    [
                        [0, 1, 1],
                    ], dtype=np.float32)
            ],
            []
        ]

        pt = ParticleTracker(frames=frames_example, time=times_example, automatic_update=False)
        pt.integration_radius_of_intensity_peaks = 1
        pt.maximum_number_of_frames_a_particle_can_disappear_and_still_be_linked_to_other_particles = 2
        pt.maximum_distance_a_particle_can_travel_between_frames = 1

        pt._particle_positions = particle_positions
        pt._calculate_particle_moments()

        pt._initialise_association_and_cost_matrices()
        pt._calculate_cost_matrices()

        for frame_index, _ in enumerate(pt._cost_matrix):
            for future_frame_index, _ in enumerate(pt._cost_matrix[frame_index]):
                np.testing.assert_array_almost_equal(pt._cost_matrix[frame_index][future_frame_index], expected_cost_matrix[frame_index][future_frame_index])

    def test_initial_links_in_association_matrix(self):
        """
        Verification of the initial links in the association matrix is set correctly.
        """
        frames_example = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
        ], dtype=np.float32)

        times_example = np.array([0, 1, 2])

        cost_matrix = [
            [
                np.array(
                    [
                        [0, 1, 1, 1],
                        [1, 0.2, 0.5, 0.3],
                        [1, 0.1, 0.6, 0.3],
                        [1, 0.2, 0.3, np.inf],
                        [1, 0.3, 0.1, 0.8]
                    ], dtype=np.float32),
                np.array(
                    [
                        [0, 4, 4],
                        [4, 5, 6],
                        [4, 7, 8],
                        [4, 9, 10]
                    ], dtype=np.float32)
            ]
        ]
        initial_association_matrix = [
            [
                np.zeros((5, 4), dtype=bool),
                np.zeros((4, 3), dtype=bool)
            ]
        ]

        expected_association_matrix = [
            [
                np.array(
                    [
                        [1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 0, 1],
                        [0, 0, 1, 0],
                        [1, 0, 0, 0]
                    ], dtype=bool),
                np.array(
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                        [1, 0, 0]
                    ], dtype=bool)

            ]
        ]

        pt = ParticleTracker(frames=frames_example, time=times_example, automatic_update=False)

        pt._cost_matrix = cost_matrix
        pt._association_matrix = initial_association_matrix
        pt._create_initial_links_in_association_matrix()

        for frame_index, _ in enumerate(pt._cost_matrix):
            for future_frame_index, _ in enumerate(pt._cost_matrix[frame_index]):
                np.testing.assert_array_almost_equal(pt._association_matrix[frame_index][future_frame_index], expected_association_matrix[frame_index][future_frame_index])

    def test_optimisation_of_association_matrix(self):
        """
        Test that the optimisation of the association matrix minimises the cost.
        """
        frames_example = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
        ], dtype=np.float32)

        times_example = np.array([0, 1, 2])

        cost_matrix = [
            [
                np.array(
                    [
                        [0, 1, 1],
                        [1, 0.8, 0.9],
                        [1, 0.1, 0.9],
                    ], dtype=np.float32),
                np.array(
                    [
                        [0, 4, 4],
                        [4, 11, 12],
                        [4, 14, 13],
                        [4, 3, 14]
                    ], dtype=np.float32),
                np.array(
                    [
                        [0, 4, 4, 4],
                        [4, 13, 14, 3],
                        [4, 14, 5, 1]
                    ], dtype=np.float32)
            ]
        ]

        initial_association_matrix = [
            [
                np.zeros((3, 3), dtype=bool),
                np.zeros((4, 3), dtype=bool),
                np.zeros((3, 4), dtype=bool)
            ]
        ]

        expected_association_matrix = [
            [
                np.array(
                    [
                        [1, 0, 0],
                        [0, 0, 1],
                        [0, 1, 0],
                    ], dtype=bool),
                np.array(
                    [
                        [1, 0, 1],
                        [1, 0, 0],
                        [1, 0, 0],
                        [0, 1, 0],
                    ], dtype=bool),
                np.array(
                    [
                        [1, 1, 0, 0],
                        [0, 0, 0, 1],
                        [0, 0, 1, 0]
                    ], dtype=bool)
            ]
        ]

        pt = ParticleTracker(frames=frames_example, time=times_example, automatic_update=False)

        pt._cost_matrix = cost_matrix
        pt._association_matrix = initial_association_matrix
        pt._create_initial_links_in_association_matrix()
        pt._optimise_association_matrix()

        for frame_index, _ in enumerate(pt._association_matrix):
            for future_frame_index, _ in enumerate(pt._association_matrix[frame_index]):
                np.testing.assert_array_almost_equal(pt._association_matrix[frame_index][future_frame_index], expected_association_matrix[frame_index][future_frame_index])

    def test_creating_trajectory(self):

        frames_example = np.array([
            [0, 0.1, 0.5],
            [0, 0.6, 0.2],
            [1, 0.1, 0.1],
            [1, 0.1, 0.1]
        ], dtype=np.float32)

        times_example = np.array([1, 2, 3, 4])

        particle_position = [
            np.array([0, 1]),
            np.array([0, 1, 2]),
            np.array([0, 1]),
            np.array([0, 1, 2]),
        ]

        zeroth_order_moments = [
            np.array([1, 2]),
            np.array([1, 2, 3]),
            np.array([1, 2]),
            np.array([1, 2, 3]),
        ]

        second_order_moments = [
            np.array([2, 3]),
            np.array([2, 3, 4]),
            np.array([2, 3]),
            np.array([2, 3, 4]),
        ]

        association_matrix = [
            [

                np.array(
                    [
                        [1, 0, 0, 1],
                        [0, 0, 1, 0],
                        [0, 1, 0, 0],
                    ], dtype=bool),
                np.array(
                    [
                        [1, 0, 0],
                        [1, 0, 0],
                        [0, 0, 1],
                    ], dtype=bool),
                np.array(
                    [
                        [1, 0, 0, 1],
                        [0, 1, 0, 0],
                        [0, 0, 1, 0]
                    ], dtype=bool)
            ],
            [
                np.array(
                    [
                        [1, 1, 1],
                        [0, 0, 1],
                        [0, 1, 0],
                        [1, 0, 0]
                    ], dtype=bool),
                np.array(
                    [
                        [1, 0, 0, 0],
                        [0, 1, 0, 0],
                        [0, 0, 1, 0],
                        [0, 0, 0, 1]
                    ], dtype=bool)
            ],
            [
                np.array(
                    [
                        [1, 0, 0, 1],
                        [0, 0, 1, 0],
                        [0, 1, 0, 0]
                    ], dtype=bool),
            ],
            []
        ]

        trajectories = [
            np.zeros((4,),
                     dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32), ('second_order_moment', np.float32)]),
            np.zeros((4,),
                     dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32), ('second_order_moment', np.float32)]),
            np.zeros((2,),
                     dtype=[('frame_index', np.int16), ('time', np.float32), ('position', np.float32), ('zeroth_order_moment', np.float32), ('second_order_moment', np.float32)])
        ]

        trajectories[0]['frame_index'] = np.array([0, 1, 2, 3])
        trajectories[0]['time'] = np.array([1, 2, 3, 4])
        trajectories[0]['position'] = np.array([0, 1, 0, 1])
        trajectories[0]['zeroth_order_moment'] = np.array([1, 2, 1, 2])
        trajectories[0]['second_order_moment'] = np.array([2, 3, 2, 3])

        trajectories[1]['frame_index'] = np.array([0, 1, 2, 3])
        trajectories[1]['time'] = np.array([1, 2, 3, 4])
        trajectories[1]['position'] = np.array([1, 0, 1, 0])
        trajectories[1]['zeroth_order_moment'] = np.array([2, 1, 2, 1])
        trajectories[1]['second_order_moment'] = np.array([3, 2, 3, 2])

        trajectories[2]['frame_index'] = np.array([1, 3])
        trajectories[2]['time'] = np.array([2, 4])
        trajectories[2]['position'] = np.array([2, 2])
        trajectories[2]['zeroth_order_moment'] = np.array([3, 3])
        trajectories[2]['second_order_moment'] = np.array([4, 4])

        pt = ParticleTracker(time=times_example, frames=frames_example, automatic_update=False)

        pt._association_matrix = association_matrix
        pt._particle_positions = particle_position
        pt._zeroth_order_moments = zeroth_order_moments
        pt._second_order_moments = second_order_moments

        pt._update_trajectories()

        for index, t in enumerate(pt.trajectories):
            np.testing.assert_array_equal(trajectories[index], t.particle_positions)


if __name__ == '__main__':
    unittest.main()
